#include <stdio.h>
#include <getopt.h>
#include <string.h>
#include <stdlib.h>
#include <stdbool.h>
#include <inttypes.h>
#include "linelet_format.h"
#include "timer.h"
#include "util.h"
#include "log.h"
#include "memory.h"

static unsigned threads;
static char *input;
static char *output;
static bool quiet;

static void print_help(void)
{
   linelet_log_printf(LINELET_LOG_INFO, "linelet decode [-q/--quiet] [-t/--threads num threads] [-o/--output raw YUV output] [-i LLV video input] [-h/--help This help]\n");
}

static void parse_input(int argc, char *argv[])
{
   threads = 1;

   const struct option opts[] = {
      { "help", 0, NULL, 'h' },
      { "input", 1, NULL, 'i' },
      { "output", 1, NULL, 'o' },
      { "threads", 1, NULL, 't' },
      { "quiet", 0, NULL, 'q' },
      { NULL },
   };

   const char *optstring = "hi:o:t:q";

   for (;;)
   {
      int c = getopt_long(argc, argv, optstring, opts, NULL);
      if (c == -1)
         break;

      switch (c)
      {
         case 'h':
            print_help();
            exit(EXIT_SUCCESS);

         case 'q':
            quiet = true;
            break;

         default:
         case '?':
            print_help();
            exit(EXIT_FAILURE);

         case 'i':
            free(input);
            input = strdup(optarg);
            break;

         case 't':
            threads = strtoul(optarg, NULL, 0);
            break;

         case 'o':
            free(output);
            output = strdup(optarg);
            break;
      }
   }

   if (optind < argc)
   {
      print_help();
      exit(EXIT_FAILURE);
   }
}

static bool check_magic(FILE *file)
{
   char magic[4];
   if (fread(&magic, sizeof(magic), 1, file) != 1)
      return false;

   return memcmp(magic, "LLV1", 4) == 0;
}

int decode_main(int argc, char *argv[])
{
   parse_input(argc, argv);

   FILE *in_file = strcmp(input, "-") ? fopen(input, "rb") : stdin;
   if (!in_file)
   {
      linelet_log_printf(LINELET_LOG_INFO, "Failed to open input file %s.\n", input);
      return EXIT_FAILURE;
   }

   FILE *out_file = strcmp(output, "-") ? fopen(output, "wb") : stdout;
   if (!out_file)
   {
      linelet_log_printf(LINELET_LOG_INFO, "Failed to open output file %s.\n", output);
      return EXIT_FAILURE;
   }

   if (!check_magic(in_file))
   {
      linelet_log_printf(LINELET_LOG_INFO, "Didn't find LLV video in input.\n");
      return EXIT_FAILURE;
   }

   unsigned total_frames = 0;
   linelet_format_decoder_t *dec = NULL;

   uint8_t *encoded_buffer[2] = {NULL};
   uint8_t *decoded_buffer[2] = {NULL};
   size_t encoded_buffer_size[2] = {0};
   size_t decoded_buffer_size[2] = {0};
   size_t decoded_buffer_current_size[2] = {0};
   unsigned buffer_ptr = 0;
   bool pending_decode = false;

   int64_t start_time = linelet_timer_get_nsec();
   int64_t current_time = start_time;

   // Do I/O while threads are crunching.
   for (;;)
   {
      // Read N
      // Wait N - 1
      // Queue N
      // Write N - 1

      uint8_t frame_size_buf[4];
      if (fread(frame_size_buf, sizeof(frame_size_buf), 1, in_file) != 1)
         break;
      uint32_t frame_size = linelet_read32_be(frame_size_buf);

      if (encoded_buffer_size[buffer_ptr] < frame_size)
      {
         linelet_free(encoded_buffer[buffer_ptr]);
         encoded_buffer[buffer_ptr] = linelet_malloc(frame_size);
         if (!encoded_buffer[buffer_ptr])
            break;
         encoded_buffer_size[buffer_ptr] = frame_size;
      }

      if (fread(encoded_buffer[buffer_ptr], 1, frame_size, in_file) != frame_size)
         break;

      struct linelet_format_decoder_frame_params params = {0};
      if (!linelet_format_decoder_parse_header(&params, encoded_buffer[buffer_ptr], frame_size))
         break;

      if (params.format == LINELET_FMT_YCBCR8_RCT_444)
         params.format = LINELET_FMT_GBRP8;
      else if (params.format == LINELET_FMT_YCBCR10_RCT_444)
         params.format = LINELET_FMT_GBRP10;

      if (!dec)
         dec = linelet_format_decoder_new(&(const struct linelet_format_decoder_options) { .format = params.format, .threads = threads });
      if (!dec)
         break;

      unsigned subsample = linelet_format_to_subsample(params.format);
      unsigned bpp = linelet_format_to_bpp(params.format);
      if (!bpp)
         break;

      size_t luma_size = params.width * params.height * bpp;
      size_t chroma_size = luma_size >> subsample;

      size_t buffer_size = luma_size + chroma_size * 2;

      if (decoded_buffer_size[buffer_ptr] < buffer_size)
      {
         linelet_free(decoded_buffer[buffer_ptr]);
         decoded_buffer[buffer_ptr] = linelet_malloc(buffer_size);
         if (!decoded_buffer[buffer_ptr])
            break;
         decoded_buffer_size[buffer_ptr] = buffer_size;
      }
      decoded_buffer_current_size[buffer_ptr] = buffer_size;

      void *dsts[] = {
         decoded_buffer[buffer_ptr],
         decoded_buffer[buffer_ptr] + luma_size,
         decoded_buffer[buffer_ptr] + luma_size + chroma_size,
      };

      const size_t strides[] = {
         params.width * bpp,
         (params.width >> subsample) * bpp,
         (params.width >> subsample) * bpp,
      };

      if (pending_decode)
      {
         if (!linelet_format_decoder_process_frame_wait(dec))
            break;
         total_frames++;
         current_time = linelet_timer_get_nsec();
      }

      if (!linelet_format_decoder_process_frame_async(dec,
               encoded_buffer[buffer_ptr], frame_size, dsts, strides))
      {
         linelet_log_printf(LINELET_LOG_INFO, "Failed to decode frame.\n");
         break;
      }

      if (pending_decode)
      {
         if (fwrite(decoded_buffer[buffer_ptr ^ 1], 1,
                  decoded_buffer_current_size[buffer_ptr ^ 1],
                  out_file) != buffer_size)
            break;
      }

      if (!quiet && total_frames)
         linelet_log_printf(LINELET_LOG_INFO, "\r[Decoder]: Decoder FPS: %8.3f, Decoding Time: %8.3f s",
               1000000000.0 * total_frames / (current_time - start_time),
               (current_time - start_time) / 1000000000.0);


      pending_decode = true;
      buffer_ptr ^= 1;
   }

   if (pending_decode)
   {
      if (linelet_format_decoder_process_frame_wait(dec))
      {
         fwrite(decoded_buffer[buffer_ptr ^ 1], 1,
                  decoded_buffer_current_size[buffer_ptr ^ 1],
                  out_file);
      }
   }

   free(input);
   free(output);
   fclose(in_file);
   fclose(out_file);
   linelet_free(encoded_buffer[0]);
   linelet_free(encoded_buffer[1]);
   linelet_free(decoded_buffer[0]);
   linelet_free(decoded_buffer[1]);
   linelet_format_decoder_free(dec);
   return 0;
}

