#include <stdio.h>
#include <getopt.h>
#include <string.h>
#include <stdlib.h>
#include <inttypes.h>
#include "linelet_format.h"
#include "image.h"
#include "memory.h"
#include "timer.h"
#include "util.h"
#include "log.h"

static enum linelet_format_format format;
static unsigned width, height;
static unsigned slice_height;
static double rate;
static unsigned threads;
static unsigned precinct;
static unsigned decomp;
static bool quiet;
static enum linelet_encoder_rate_allocation_style style = LINELET_ENCODER_RATE_ALLOCATION_CBR;
static unsigned quant;
static unsigned filter;
static struct linelet_format_ratio fps;

static char *input;
static char *output;

static void print_help(void)
{
   linelet_log_printf(LINELET_LOG_INFO, "linelet encode [-q/--quiet] [-Q/--quant VBR quant] [-S/--slice-height slice height] [-s/--size widthxheight] [-f/--pix_fmt format] [-m/--colormatrix <bt601|bt709 (default)>] [--filter apply prefilter/postfilter] [-t/--threads num threads] [--fps num/den or num] [-r/--rate rate (0 to 1)] [-o/--output LLV video output] [-i/--input raw YUV input] [-h/--help This help]\n");
   linelet_log_printf(LINELET_LOG_INFO, "\tFormats: yuv444p yuv422p yuv444p10 yuv422p10 yuv444p12 yuv422p12 gbrp gbrp10\n");
}

static void parse_input(int argc, char *argv[])
{
   threads = 1;
   slice_height = 8;
   rate = 0.5;
   precinct = 128;
   decomp = 5;

   const struct option opts[] = {
      { "help", 0, NULL, 'h' },
      { "pix_fmt", 1, NULL, 'f' },
      { "colormatrix", 1, NULL, 'm' },
      { "size", 1, NULL, 's' },
      { "input", 1, NULL, 'i' },
      { "output", 1, NULL, 'o' },
      { "slice_height", 1, NULL, 'S' },
      { "threads", 1, NULL, 't' },
      { "rate", 1, NULL, 'r' },
      { "quiet", 0, NULL, 'q' },
      { "quant", 1, NULL, 'Q' },
      { "precinct", 1, NULL, 'P' },
      { "decompositions", 1, NULL, 'D' },
      { "filter", 1, NULL, 'F' },
      { "fps", 1, NULL, 'R' },
      { NULL },
   };

   const char *optstring = "hf:m:s:i:o:S:t:r:qQ:P:D:";

   char *format_str = NULL;
   char *format_matrix = NULL;

   for (;;)
   {
      int c = getopt_long(argc, argv, optstring, opts, NULL);
      if (c == -1)
         break;

      switch (c)
      {
         case 'h':
            print_help();
            exit(EXIT_SUCCESS);

         case 'q':
            quiet = true;
            break;

         case 'R':
         {
            int ret = sscanf(optarg, "%u/%u", &fps.num, &fps.den);
            if (ret == 1)
               fps.den = 1;
            else if (ret != 2)
            {
               print_help();
               exit(EXIT_SUCCESS);
            }
            break;
         }

         default:
         case '?':
            print_help();
            exit(EXIT_FAILURE);

         case 'F':
            filter = strtoul(optarg, NULL, 0);
            break;

         case 'D':
            decomp = strtoul(optarg, NULL, 0);
            break;

         case 'P':
            precinct = strtoul(optarg, NULL, 0);
            break;

         case 'r':
            rate = strtod(optarg, NULL);
            break;

         case 'f':
            free(format_str);
            format_str = strdup(optarg);
            break;

         case 'Q':
            quant = strtoul(optarg, NULL, 0);
            style = LINELET_ENCODER_RATE_ALLOCATION_MINIMUM_QUANT;
            break;

         case 'm':
            free(format_matrix);
            format_matrix = strdup(optarg);
            break;

         case 's':
            if (sscanf(optarg, "%ux%u", &width, &height) != 2)
            {
               print_help();
               exit(EXIT_FAILURE);
            }
            break;

         case 'S':
            slice_height = strtoul(optarg, NULL, 0);
            break;

         case 't':
            threads = strtoul(optarg, NULL, 0);
            break;

         case 'i':
            free(input);
            input = strdup(optarg);
            break;

         case 'o':
            free(output);
            output = strdup(optarg);
            break;
      }
   }

   if (optind < argc)
   {
      print_help();
      exit(EXIT_FAILURE);
   }

   format = linelet_format_parse(format_str, format_matrix);
   if (format == LINELET_FMT_UNKNOWN)
   {
      print_help();
      exit(EXIT_FAILURE);
   }

   free(format_str);
   free(format_matrix);

   if (!input)
   {
      linelet_log_printf(LINELET_LOG_INFO, "Missing input file.\n");
      print_help();
      exit(EXIT_FAILURE);
   }
   else if (!output)
   {
      linelet_log_printf(LINELET_LOG_INFO, "Missing output file.\n");
      print_help();
      exit(EXIT_FAILURE);
   }
   else if (!width || !height)
   {
      linelet_log_printf(LINELET_LOG_INFO, "Missing --size.\n");
      print_help();
      exit(EXIT_FAILURE);
   }
   else if (format == LINELET_FMT_UNKNOWN)
   {
      linelet_log_printf(LINELET_LOG_INFO, "Missing input pixel format --pix_fmt.\n");
      print_help();
      exit(EXIT_FAILURE);
   }
}

int encode_main(int argc, char *argv[])
{
   parse_input(argc, argv);

   FILE *in_file = strcmp(input, "-") ? fopen(input, "rb") : stdin;
   if (!in_file)
   {
      linelet_log_printf(LINELET_LOG_INFO, "Failed to open input file %s.\n", input);
      return EXIT_FAILURE;
   }

   FILE *out_file = strcmp(output, "-") ? fopen(output, "wb") : stdout;
   if (!out_file)
   {
      linelet_log_printf(LINELET_LOG_INFO, "Failed to open output file %s.\n", output);
      return EXIT_FAILURE;
   }

   char magic[] = "LLV1";
   if (fwrite(magic, 4, 1, out_file) != 1)
      linelet_log_printf(LINELET_LOG_INFO, "Failed to write! (Disk full?)\n");

   linelet_format_encoder_t *enc = linelet_format_encoder_new(
         &(const struct linelet_format_encoder_options) {
            .format = format,
            .width = width,
            .height = height,
            .precinct_samples = precinct,
            .decomposition_levels = decomp,
            .slice_height = slice_height,
            .threads = threads,
            .rate_control = style,
            .q = quant,
            .filter = filter,
            .frame_rate = fps,
         });

   if (!enc)
   {
      linelet_log_printf(LINELET_LOG_INFO, "Failed to init encoder.\n");
      return EXIT_FAILURE;
   }

   unsigned subsample = linelet_format_to_subsample(format);
   unsigned bpp = linelet_format_to_bpp(format);
   unsigned bits = linelet_format_to_bits(format) * (subsample ? 2 : 3);

   size_t luma_size = width * height * bpp;
   size_t chroma_size = luma_size >> subsample;

   unsigned frame_size_raw = width * height * bits / 8;
   unsigned frame_size = rate * frame_size_raw;
   unsigned input_size = luma_size + chroma_size * 2;

   // Double-buffer to allow async encoding.
   uint8_t *encoder_buffer = linelet_malloc(2 * frame_size);
   uint8_t *input_buf = linelet_malloc(2 * input_size);
   unsigned buffer_ptr = 0;

   if (!input_buf || !encoder_buffer)
   {
      linelet_log_printf(LINELET_LOG_INFO, "Failed to allocate buffers.\n");
      return EXIT_FAILURE;
   }

   if (!quiet)
      linelet_log_printf(LINELET_LOG_INFO, "[Encoder]: Allocating %u bytes per frame.\n",
            frame_size);

   const void *bufs[2][3] = {
      { input_buf, input_buf + luma_size, input_buf + luma_size + chroma_size },
      { input_buf + input_size, input_buf + input_size + luma_size, input_buf + input_size + luma_size + chroma_size },
   };

   const size_t strides[] = { width * bpp, (width >> subsample) * bpp, (width >> subsample) * bpp };

   unsigned total_frames = 0;
   uint64_t total_size = 0;
   uint64_t total_in_size = 0;
   bool pending_encode = false;

   int64_t start_time = linelet_timer_get_nsec();
   int64_t current_time = start_time;

   // Do I/O while threads are crunching.
   for (;;)
   {
      // Read N
      // Wait N - 1
      // Queue N
      // Write N - 1

      uint8_t *input = input_buf + buffer_ptr * input_size;
      uint8_t *output = encoder_buffer + buffer_ptr * frame_size;

      uint8_t *prev_output = encoder_buffer + (buffer_ptr ^ 1) * frame_size;

      if (fread(input, 1, input_size, in_file) != input_size)
         break;

      total_in_size += frame_size_raw;

      size_t encoded = 0;
      if (pending_encode)
      {
         encoded = linelet_format_encoder_process_frame_wait(enc);
         total_frames++;
         current_time = linelet_timer_get_nsec();
         total_size += encoded;
      }

      linelet_format_encoder_process_frame_async(enc, bufs[buffer_ptr], strides,
            output, frame_size);

      // Flush out encoded data.
      if (pending_encode)
      {
         uint8_t frame_size_buf[4];
         linelet_write32_be(frame_size_buf, encoded);
         if (fwrite(frame_size_buf, sizeof(frame_size_buf), 1, out_file) != 1)
            linelet_log_printf(LINELET_LOG_INFO, "Failed to write! (Disk full?)\n");
         if (fwrite(prev_output, 1, encoded, out_file) != encoded)
            break;
      }

      if (!quiet && total_frames)
      {
         linelet_log_printf(LINELET_LOG_INFO, "\r[Encoder]: FPS: %8.3f, Encoding Time: %8.3f s, Encoded size: %10.3f MiB (%.3f %%)",
               1000000000.0 * total_frames / (current_time - start_time),
               (current_time - start_time) / 1000000000.0,
               total_size / (1024.0 * 1024.0), 100.0 * total_size / total_in_size);
      }

      buffer_ptr ^= 1;
      pending_encode = true;
   }

   if (pending_encode)
   {
      size_t encoded = linelet_format_encoder_process_frame_wait(enc);
      uint8_t *prev_output = encoder_buffer + (buffer_ptr ^ 1) * frame_size;
      uint8_t frame_size_buf[4];
      linelet_write32_be(frame_size_buf, encoded);
      if (fwrite(frame_size_buf, sizeof(frame_size_buf), 1, out_file) != 1)
         linelet_log_printf(LINELET_LOG_INFO, "Failed to write! (Disk full?)\n");
      if (fwrite(prev_output, 1, encoded, out_file) != encoded)
         linelet_log_printf(LINELET_LOG_INFO, "Failed to write! (Disk full?)\n");
   }

   if (!quiet)
      linelet_log_printf(LINELET_LOG_INFO, "\n");

   free(input);
   free(output);
   linelet_format_encoder_free(enc);
   linelet_free(input_buf);
   linelet_free(encoder_buffer);
   if (in_file && in_file != stdin)
      fclose(in_file);
   if (out_file && out_file != stdout)
      fclose(out_file);

   return 0;
}

