#include "image.h"
#include "memory.h"
#include "linelet_format.h"
#include "log.h"
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <getopt.h>
#include <string.h>
#include <stdbool.h>

static unsigned width, height;
static enum linelet_format_format format;
static bool total;
static unsigned slice_height;

static void print_help(void)
{
   linelet_log_printf(LINELET_LOG_INFO, "linelet psnr_video <raw YUV a> <raw YUV b> [-s/--size widthxheight] [-f/--pix_fmt format] [-h/--help This help] [-t/--total only report total]\n");
}

static void parse_input(int argc, char *argv[])
{
   const struct option opts[] = {
      { "help", 0, NULL, 'h' },
      { "pix_fmt", 1, NULL, 'f' },
      { "size", 1, NULL, 's' },
      { "total", 0, NULL, 't' },
      { "slice-height", 1, NULL, 'S' },
      { NULL },
   };

   const char *optstring = "hf:s:tS:";

   for (;;)
   {
      int c = getopt_long(argc, argv, optstring, opts, NULL);
      if (c == -1)
         break;

      switch (c)
      {
         case 'h':
            print_help();
            exit(EXIT_SUCCESS);

         default:
         case '?':
            print_help();
            exit(EXIT_FAILURE);

         case 'S':
            slice_height = strtoul(optarg, NULL, 0);
            break;

         case 't':
            total = true;
            break;

         case 'f':
            format = linelet_format_parse(optarg, NULL);
            break;

         case 's':
            if (sscanf(optarg, "%ux%u", &width, &height) != 2)
            {
               print_help();
               exit(EXIT_FAILURE);
            }
            break;
      }
   }

   if (optind != argc - 2)
   {
      print_help();
      exit(EXIT_FAILURE);
   }

   if (format == LINELET_FMT_UNKNOWN)
   {
      print_help();
      exit(EXIT_FAILURE);
   }
   else if (!width || !height)
   {
      linelet_log_printf(LINELET_LOG_INFO, "Missing --size.\n");
      print_help();
      exit(EXIT_FAILURE);
   }
   else if (format == LINELET_FMT_UNKNOWN)
   {
      linelet_log_printf(LINELET_LOG_INFO, "Missing input pixel format --pix_fmt.\n");
      print_help();
      exit(EXIT_FAILURE);
   }
}

int psnr_video_main(int argc, char *argv[])
{
   parse_input(argc, argv);

   FILE *a_file = strcmp(argv[optind + 0], "-") ? fopen(argv[optind + 0], "rb") : stdin;
   FILE *b_file = strcmp(argv[optind + 1], "-") ? fopen(argv[optind + 1], "rb") : stdin;
   if (!a_file || !b_file)
   {
      linelet_log_printf(LINELET_LOG_INFO, "Failed to open files.\n");
      return EXIT_FAILURE;
   }

   unsigned subsample = linelet_format_to_subsample(format);
   unsigned bpp = linelet_format_to_bpp(format);
   unsigned bits = linelet_format_to_bits(format);

   size_t luma_size = width * height * bpp;
   size_t chroma_size = luma_size >> subsample;

   void *luma_buf_a = linelet_malloc(luma_size);
   void *cb_buf_a = linelet_malloc(chroma_size);
   void *cr_buf_a = linelet_malloc(chroma_size);
   void *luma_buf_b = linelet_malloc(luma_size);
   void *cb_buf_b = linelet_malloc(chroma_size);
   void *cr_buf_b = linelet_malloc(chroma_size);

   double mse_luma_total = 0.0;
   double mse_cb_total = 0.0;
   double mse_cr_total = 0.0;
   unsigned mse_frames = 0;

   double mse_luma_lines[LINELET_MAX_LINES] = {0.0};
   unsigned mse_lines[LINELET_MAX_LINES] = {0};

   for (;;)
   {
      if (fread(luma_buf_a, 1, luma_size, a_file) != luma_size)
         break;
      if (fread(cb_buf_a, 1, chroma_size, a_file) != chroma_size)
         break;
      if (fread(cr_buf_a, 1, chroma_size, a_file) != chroma_size)
         break;
      if (fread(luma_buf_b, 1, luma_size, b_file) != luma_size)
         break;
      if (fread(cb_buf_b, 1, chroma_size, b_file) != chroma_size)
         break;
      if (fread(cr_buf_b, 1, chroma_size, b_file) != chroma_size)
         break;

      double mse_luma = 0.0, mse_cb, mse_cr;
      if (bpp == 2)
      {
         for (unsigned i = 0; i < height; i++)
         {
            double mse = linelet_image_component_mse_16((uint16_t*)luma_buf_a + width * i, (uint16_t*)luma_buf_b + width * i, width);
            mse_luma += mse;
            if (slice_height)
            {
               unsigned line = i & (slice_height - 1);
               mse_luma_lines[line] += mse;
               mse_lines[line]++;
            }
         }

         mse_cb = linelet_image_component_mse_16(cb_buf_a, cb_buf_b, (width * height) >> subsample);
         mse_cr = linelet_image_component_mse_16(cr_buf_a, cr_buf_b, (width * height) >> subsample);
      }
      else
      {
         for (unsigned i = 0; i < height; i++)
         {
            double mse = linelet_image_component_mse_8((uint8_t*)luma_buf_a + width * i, (uint8_t*)luma_buf_b + width * i, width);
            mse_luma += mse;
            if (slice_height)
            {
               unsigned line = i & (slice_height - 1);
               mse_luma_lines[line] += mse;
               mse_lines[line]++;
            }
         }

         mse_cb = linelet_image_component_mse_8(cb_buf_a, cb_buf_b, (width * height) >> subsample);
         mse_cr = linelet_image_component_mse_8(cr_buf_a, cr_buf_b, (width * height) >> subsample);
      }

      mse_luma /= height;

      mse_luma_total += mse_luma;
      mse_cb_total += mse_cb;
      mse_cr_total += mse_cr;
      mse_frames++;

      // Not technically correct for TV range, but it's just a relative measurement, and all comparisons go through the same assumptions anyways.
      double range = (1 << bits) - 1;
      double psnr_luma = 10.0 * log10(range * range / mse_luma);
      double psnr_cb = 10.0 * log10(range * range / mse_cb);
      double psnr_cr = 10.0 * log10(range * range / mse_cr);

      if (!total)
         printf("[PSNR]: Y: %10.6f, Cb: %10.6f, Cr: %10.6f\n",
               psnr_luma, psnr_cb, psnr_cr);
   }

   if (mse_frames)
   {
      mse_luma_total /= mse_frames;
      mse_cb_total /= mse_frames;
      mse_cr_total /= mse_frames;
      double range = (1 << bits) - 1;
      double psnr_luma = 10.0 * log10(range * range / mse_luma_total);
      double psnr_cb = 10.0 * log10(range * range / mse_cb_total);
      double psnr_cr = 10.0 * log10(range * range / mse_cr_total);
      double psnr_avg = 10.0 * log10(3 * range * range / (mse_luma_total + mse_cb_total + mse_cr_total));
      printf("[PSNR Total]: Avg: %10.6f, Y: %10.6f, Cb: %10.6f, Cr: %10.6f\n",
            psnr_avg, psnr_luma, psnr_cb, psnr_cr);

      if (slice_height)
      {
         for (unsigned i = 0; i < slice_height; i++)
         {
            mse_luma_lines[i] /= mse_lines[i];
            double psnr_luma = 10.0 * log10(range * range / mse_luma_lines[i]);
            printf("[PSNR Total Line %u]: Y: %10.6f\n", i, psnr_luma);
         }
      }
   }

   if (a_file != stdin)
      fclose(a_file);
   if (b_file != stdin)
      fclose(b_file);

   return 0;
}
