#include <stdio.h>
#include <getopt.h>
#include <string.h>
#include <stdlib.h>
#include <inttypes.h>
#include <errno.h>
#include <pthread.h>
#include <signal.h>
#include "timer.h"
#include "util.h"
#include "log.h"

#ifdef _WIN32
// Win32 headers.
#define WIN32_LEAN_AND_MEAN
#include <winsock2.h>
#include <windows.h>
#include <ws2tcpip.h>
#define close(x) closesocket(x)
#define SHUT_RD SD_RECEIVE
#define SHUT_WR SD_SEND
#else
// POSIX headers.
#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>
#include <unistd.h>
#include <sys/fcntl.h>
#include <netinet/in.h>
#endif

static unsigned port;
static int buffer;
static unsigned delay;
static bool ipv6;
static char *bind_addr;
static char *multicast_addr;

struct thread_info
{
   int fd;
   pthread_mutex_t lock;
   pthread_cond_t cond;
   pthread_t thread;

   size_t packet_max_size;

   struct udp_packet *packets;
   unsigned num_packets;

   unsigned read_ptr;
   unsigned write_ptr;
   unsigned packets_count;

   bool die;
};

static struct thread_info info;

static volatile sig_atomic_t alive = 1;

#ifndef _WIN32
static void sighandler(int sig)
{
   (void)sig;
   alive = 0;
   pthread_mutex_lock(&info.lock);
   info.die = true;
   pthread_cond_signal(&info.cond);
   pthread_mutex_unlock(&info.lock);
}
#endif

static void print_help(void)
{
   linelet_log_printf(LINELET_LOG_INFO,
         "rtp_recv <out file> [-h/--help this help] [-B/--bind bind addr] [--ipv6] [--multicast] [-d/--delay max delay (packets)] [-b/--buffer UDP buffer size] [-p/--port UDP port]\n");
}

static void parse_input(int argc, char *argv[])
{
   delay = 256;
   const struct option opts[] = {
      { "help", 0, NULL, 'h' },
      { "port", 1, NULL, 'p' },
      { "buffer", 1, NULL, 'b' },
      { "delay", 1, NULL, 'd' },
      { "ipv6", 0, NULL, 'i' },
      { "multicast", 1, NULL, 'm' },
      { "bind", 1, NULL, 'B' },
      { NULL },
   };

   const char *optstring = "hp:b:d:B:m";

   for (;;)
   {
      int c = getopt_long(argc, argv, optstring, opts, NULL);
      if (c == -1)
         break;

      switch (c)
      {
         case 'h':
            print_help();
            exit(EXIT_SUCCESS);

         default:
         case '?':
            print_help();
            exit(EXIT_FAILURE);

         case 'i':
            ipv6 = true;
            break;

         case 'm':
            free(multicast_addr);
            multicast_addr = strdup(optarg);
            break;

         case 'B':
            free(bind_addr);
            bind_addr = strdup(optarg);
            break;

         case 'd':
            delay = strtoul(optarg, NULL, 0);
            break;

         case 'p':
            port = strtoul(optarg, NULL, 0);
            break;

         case 'b':
            buffer = strtol(optarg, NULL, 0);
            break;
      }
   }

   if (optind >= argc)
   {
      print_help();
      exit(EXIT_FAILURE);
   }

   if (!port)
   {
      print_help();
      exit(EXIT_FAILURE);
   }
}

static int open_udp_socket(const char *addr, const char *multi, unsigned port)
{
#ifdef _WIN32
   WSADATA wsa;
   if (WSAStartup(MAKEWORD(2, 2), &wsa) != 0)
   {
      WSACleanup();
      return -1;
   }
#endif

   int sock = -1;
   char port_str[32];
   snprintf(port_str, sizeof(port_str), "%u", port);
   struct addrinfo *res = NULL, *multires = NULL;

   const struct addrinfo hints = {
      .ai_family = ipv6 ? AF_INET6 : AF_INET,
      .ai_socktype = SOCK_DGRAM,
      .ai_flags = AI_PASSIVE,
   };

   int ret = getaddrinfo(addr, port_str, &hints, &res);
   if (ret < 0)
   {
      linelet_log_printf(LINELET_LOG_ERROR, "getaddrinfo(): %s\n", gai_strerror(ret));
      goto error;
   }

   sock = socket(res->ai_family, res->ai_socktype, res->ai_protocol);
   if (sock < 0)
   {
      perror("socket");
      goto error;
   }

   int yes = 1;
   if (setsockopt(sock, SOL_SOCKET, SO_REUSEADDR, (const void*)&yes, sizeof(yes)) < 0)
   {
      perror("setsockopt");
      goto error;
   }

   ret = bind(sock, res->ai_addr, res->ai_addrlen);
   if (ret < 0)
   {
      perror("bind");
      goto error;
   }

   if (multi)
   {
      ret = getaddrinfo(multi, port_str, &hints, &multires);
      if (ret < 0)
      {
         linelet_log_printf(LINELET_LOG_ERROR, "getaddrinfo(): %s\n", gai_strerror(ret));
         goto error;
      }

      if (ipv6)
      {
         const struct sockaddr_in6 *sin = (const struct sockaddr_in6*)multires->ai_addr;
         struct ipv6_mreq mreq = {
            .ipv6mr_multiaddr = sin->sin6_addr,
         };

         if (setsockopt(sock, IPPROTO_IPV6, IPV6_ADD_MEMBERSHIP, (const void*)&mreq, sizeof(mreq)) < 0)
         {
            perror("setsockopt - multicast");
            goto error;
         }
      }
      else
      {
         const struct sockaddr_in *sin = (const struct sockaddr_in*)multires->ai_addr;
         struct ip_mreq mreq;
         mreq.imr_multiaddr = sin->sin_addr;
         mreq.imr_interface.s_addr = htonl(INADDR_ANY);
         if (setsockopt(sock, IPPROTO_IP, IP_ADD_MEMBERSHIP, (const void*)&mreq, sizeof(mreq)) < 0)
         {
            perror("setsockopt - multicast");
            goto error;
         }
      }
   }

   if (buffer)
   {
      if (setsockopt(sock, SOL_SOCKET, SO_SNDBUF, (const void*)&buffer, sizeof(int)) < 0)
         linelet_log_printf(LINELET_LOG_WARN, "Failed to set UDP buffer size.\n");
   }

   int size = 0;
   getsockopt(sock, SOL_SOCKET, SO_SNDBUF, (void*)&size, (socklen_t[]) { sizeof(int) });
   linelet_log_printf(LINELET_LOG_INFO, "UDP buffer size: %d\n", size);

   freeaddrinfo(res);
   if (multires)
      freeaddrinfo(multires);
   return sock;

error:
   if (sock >= 0)
      close(sock);
   if (res)
      freeaddrinfo(res);
   if (multires)
      freeaddrinfo(multires);
   return -1;
}

#define TS_PACKET_SIZE 188
#define TS_PACKETS 7
#define RTP_PAYLOAD_SIZE (TS_PACKET_SIZE * TS_PACKETS)
#define RTP_HEADER_SIZE 12
#define RTP_PACKET_SIZE (RTP_PAYLOAD_SIZE + RTP_HEADER_SIZE)
#define RTP_VERSION 2
#define RTP_PAYLOAD_TYPE 33 // MPEG2-TS

struct rtp_payload
{
   uint8_t *data;
   uint16_t size;
   uint16_t seq;
};

struct rtp_header 
{
   uint8_t version : 2;
   uint8_t padding : 1;
   uint8_t extension : 1;
   uint8_t csrc : 4;
   uint8_t marker : 1;
   uint8_t payload_type : 7;
   uint16_t seq;
   uint32_t timestamp;
   uint32_t ssrc;
   uint32_t csrcs[16];
};

bool rtp_header_parse(struct rtp_header *header, const uint8_t **payload, size_t *size)
{
   const uint8_t *p = *payload;
   header->version = p[0] >> 6;
   header->padding = !!(p[0] & 0x20);
   header->extension = !!(p[0] & 0x10);
   header->csrc = p[0] & 0x0f;
   header->marker = !!(p[1] & 0x80);
   header->payload_type = p[1] & 0x7f;
   header->seq = linelet_read16_be(p + 2);
   header->timestamp = linelet_read32_be(p + 4);
   header->ssrc = linelet_read32_be(p + 8);

   p += 12;
   if (*size < 12 + header->csrc * 4)
      return false;

   for (unsigned i = 0; i < header->csrc; i++, p += 4)
      header->csrcs[i] = linelet_read32_be(p);

   *size -= p - *payload;
   *payload = p;
   return true;
}

typedef struct rtp_sliding_window
{
   struct rtp_payload *payloads;
   unsigned num_payloads;
   unsigned cur_payloads;

   int next_seq;
} rtp_sliding_window_t;

typedef void (*rtp_sliding_window_push_cb)(const uint8_t *payload, size_t size, void *userdata);

void rtp_sliding_window_free(rtp_sliding_window_t *rtp)
{
   if (!rtp)
      return;

   for (unsigned p = 0; p < rtp->num_payloads; p++)
      free(rtp->payloads[p].data);
   free(rtp->payloads);
   free(rtp);
}

rtp_sliding_window_t *rtp_sliding_window_new(unsigned payloads, size_t max_payload_size)
{
   rtp_sliding_window_t *rtp = calloc(1, sizeof(*rtp));
   if (!rtp)
      return NULL;

   rtp->payloads = calloc(payloads, sizeof(*rtp->payloads));
   if (!rtp->payloads)
      goto error;
   rtp->num_payloads = payloads;

   for (unsigned p = 0; p < payloads; p++)
      if (!(rtp->payloads[p].data = malloc(max_payload_size)))
         goto error;

   rtp->next_seq = -1;
   return rtp;

error:
   rtp_sliding_window_free(rtp);
   return NULL;
}

// Takes care of sequence wrapping, but has to assume that values are fairly close.
static int payload_cmp(const void *a_, const void *b_)
{
   const struct rtp_payload *a = a_;
   const struct rtp_payload *b = b_;

   if (a->seq >= 0xc000 && b->seq < 0x4000)
      return -1;
   else if (a->seq < 0x4000 && b->seq >= 0xc000)
      return 1;
   else
      return a->seq - b->seq;
}

void rtp_sliding_window_push(rtp_sliding_window_t *rtp, const uint8_t *payload, size_t size, uint16_t seq,
      rtp_sliding_window_push_cb cb, void *userdata)
{
   // First packet is always expected.
   if (rtp->next_seq < 0)
      rtp->next_seq = seq;

   //linelet_log_printf(LINELET_LOG_INFO, "RTP: SEQ: %d\n", seq);

   // Arrived late. But output it anyways to keep bitrate as correct as possible.
   if ((seq - rtp->next_seq) & 0x8000)
      cb(payload, size, userdata);

   memcpy(rtp->payloads[rtp->cur_payloads].data, payload, size);
   rtp->payloads[rtp->cur_payloads].size = size;
   rtp->payloads[rtp->cur_payloads].seq = seq;
   rtp->cur_payloads++;

   // Sort payloads.
   qsort(rtp->payloads, rtp->cur_payloads, sizeof(*rtp->payloads), payload_cmp);

   unsigned payloads;
   for (;;)
   {
      // Retire all "correct" packets we can.
      for (payloads = 0;
            payloads < rtp->cur_payloads && rtp->payloads[payloads].seq == rtp->next_seq;
            payloads++)
      {
         //linelet_log_printf(LINELET_LOG_INFO, "Retiring SEQ: %d\n", rtp->payloads[payloads].seq);
         cb(rtp->payloads[payloads].data, rtp->payloads[payloads].size, userdata);
         rtp->next_seq = (rtp->next_seq + 1) & 0xffff;
      }

      // We're good for now.
      if (payloads || rtp->cur_payloads < rtp->num_payloads)
         break;

      // We've filled our sliding window buffer, and something has to be done.
      // We can assume that the packet we waited for has been lost.
      linelet_log_printf(LINELET_LOG_WARN, "RTP: Dropped packet SEQ: %d, incrementing to: %d.\n",
            rtp->next_seq, rtp->payloads[0].seq);

      rtp->next_seq = rtp->payloads[0].seq;
   }

   // Shuffle out used packets.
   for (unsigned p = payloads; p < rtp->cur_payloads; p++)
   {
      struct rtp_payload tmp = rtp->payloads[p - payloads];
      rtp->payloads[p - payloads] = rtp->payloads[p];
      rtp->payloads[p] = tmp;
   }
   rtp->cur_payloads -= payloads;
}

static void push_cb(const uint8_t *payload, size_t size, void *userdata)
{
   fwrite(payload, 1, size, userdata);
}

static void parse_rtp(rtp_sliding_window_t *rtp, FILE *file, const uint8_t *payload, size_t size)
{
   if (size < 12)
   {
      linelet_log_printf(LINELET_LOG_ERROR, "RTP: packet too small.\n");
      return;
   }

   struct rtp_header header = {0};
   if (!rtp_header_parse(&header, &payload, &size))
   {
      linelet_log_printf(LINELET_LOG_ERROR, "RTP: failed to parse header.\n");
      return;
   }

   if (header.version != RTP_VERSION)
   {
      linelet_log_printf(LINELET_LOG_ERROR, "RTP: invalid version.\n");
      return;
   }

   if (header.payload_type != RTP_PAYLOAD_TYPE)
   {
      linelet_log_printf(LINELET_LOG_ERROR, "RTP: wrong payload type.\n");
      return;
   }

   // Disallowed by SMPTE spec.
   if (header.padding || header.extension || header.marker)
   {
      linelet_log_printf(LINELET_LOG_ERROR, "RTP: padding/extension/marker used.\n");
      return;
   }

   if (size % TS_PACKET_SIZE)
   {
      linelet_log_printf(LINELET_LOG_ERROR, "RTP: TS packet size must be %d.\n", TS_PACKET_SIZE);
      return;
   }

   // SMPTE allows 1, 4 or 7 packets.
   unsigned payloads = size / 188;
   if (payloads != 1 && payloads != 4 && payloads != 7)
   {
      linelet_log_printf(LINELET_LOG_ERROR, "RTP: only 1, 4, 7 TS packets allowed in SMPTE.\n");
      return;
   }

   rtp_sliding_window_push(rtp, payload, size, header.seq, push_cb, file);
}

struct udp_packet
{
   uint8_t *data;
   size_t size;
};

void *reader_thread(void *data)
{
   struct thread_info *thr = data;
   shutdown(thr->fd, SHUT_WR); // We'll never write.

   for (;;)
   {
      pthread_mutex_lock(&thr->lock);
      while (!thr->die && thr->packets_count == thr->num_packets)
         pthread_cond_wait(&thr->cond, &thr->lock);
      pthread_mutex_unlock(&thr->lock);

      if (thr->die)
         break;

      struct udp_packet *packet = &thr->packets[thr->write_ptr];

      int ret = recvfrom(thr->fd, (void*)packet->data, thr->packet_max_size, 0, NULL, NULL);
      if (ret < 0)
         break;
      packet->size = ret;

      pthread_mutex_lock(&thr->lock);
      thr->packets_count++;
      thr->write_ptr = (thr->write_ptr + 1) % thr->num_packets;
      pthread_cond_signal(&thr->cond);
      pthread_mutex_unlock(&thr->lock);
   }

   pthread_mutex_lock(&thr->lock);
   thr->die = true;
   pthread_cond_signal(&thr->cond);
   pthread_mutex_unlock(&thr->lock);

   return NULL;
}

bool reader_thread_init(struct thread_info *info, int sock, unsigned packets, size_t packet_size)
{
   // Be a bit sloppy with error handling, but that's fine for now.
   memset(info, 0, sizeof(*info));
   info->fd = sock;

   info->packets = calloc(packets, sizeof(*info->packets));
   if (!info->packets)
      return false;

   info->num_packets = packets;
   info->packet_max_size = packet_size;
   for (unsigned i = 0; i < packets; i++)
      if (!(info->packets[i].data = malloc(packet_size)))
         return false;

   if (pthread_mutex_init(&info->lock, NULL) < 0)
      return false;
   if (pthread_cond_init(&info->cond, NULL) < 0)
      return false;

   if (pthread_create(&info->thread, NULL, reader_thread, info) < 0)
      return false;

   return true;
}

void reader_thread_free(struct thread_info *info)
{
   shutdown(info->fd, SHUT_RD); // Unblocks recvfrom() if we're blocking there.
   pthread_mutex_lock(&info->lock);
   info->die = true;
   pthread_cond_signal(&info->cond);
   pthread_mutex_unlock(&info->lock);

   pthread_join(info->thread, NULL);

   for (unsigned i = 0; i < info->num_packets; i++)
      free(info->packets[i].data);
   free(info->packets);

   pthread_mutex_destroy(&info->lock);
   pthread_cond_destroy(&info->cond);
   close(info->fd);
}

int rtp_recv_main(int argc, char *argv[])
{
   parse_input(argc, argv);

   int sock = open_udp_socket(bind_addr, multicast_addr, port);
   free(bind_addr);
   free(multicast_addr);
   if (sock < 0)
   {
      linelet_log_printf(LINELET_LOG_ERROR, "Failed to open socket.\n");
      return EXIT_FAILURE;
   }

   if (!reader_thread_init(&info, sock, delay, RTP_PACKET_SIZE))
      return EXIT_FAILURE;

   FILE *file = strcmp(argv[optind], "-") ? fopen(argv[optind], "wb") : stdout;
   if (!file)
   {
      linelet_log_printf(LINELET_LOG_ERROR, "Failed to open file.\n");
      return EXIT_FAILURE;
   }

   rtp_sliding_window_t *rtp = rtp_sliding_window_new(delay, RTP_PACKET_SIZE);
   if (!rtp)
   {
      linelet_log_printf(LINELET_LOG_ERROR, "Failed to create sliding window.\n");
      return EXIT_FAILURE;
   }

#ifndef _WIN32
   struct sigaction sa = {{0}};
   sa.sa_handler = sighandler;
   sa.sa_flags   = SA_RESTART;
   sigemptyset(&sa.sa_mask);
   sigaction(SIGINT, &sa, NULL);
   sigaction(SIGTERM, &sa, NULL);
#endif

   while (alive)
   {
      pthread_mutex_lock(&info.lock);
      while (!info.die && !info.packets_count)
         pthread_cond_wait(&info.cond, &info.lock);
      pthread_mutex_unlock(&info.lock);

      if (info.die)
         break;

      const struct udp_packet *packet = &info.packets[info.read_ptr];
      parse_rtp(rtp, file, packet->data, packet->size);
      info.read_ptr = (info.read_ptr + 1) % info.num_packets;

      pthread_mutex_lock(&info.lock);
      info.packets_count--;
      pthread_cond_signal(&info.cond);
      pthread_mutex_unlock(&info.lock);
   }

   linelet_log_printf(LINELET_LOG_INFO, "Exiting ...\n");

   reader_thread_free(&info);

   if (file && file != stdout)
      fclose(file);
   rtp_sliding_window_free(rtp);
   return 0;
}

