#ifndef LINELET_FORMAT_H__
#define LINELET_FORMAT_H__

#include <stdint.h>
#include <stddef.h>
#include <limits.h>
#include "linelet.h"

#ifdef __cplusplus
extern "C" {
#else
#include <stdbool.h>
#endif

// High-level wrapper which can encode multithreaded and dump full payloads with headers, magic, etc.
enum linelet_yuv_transform
{
   LINELET_YUV_UNKNOWN = 0,
   LINELET_YUV_RCT,
   LINELET_YUV_BT709,
   LINELET_YUV_BT601,

   LINELET_YUV_TRANSFORM_DUMMY = INT_MAX
};

enum linelet_format_format
{
   LINELET_FMT_UNKNOWN = 0,

   // RCT formats are typically only used as a stop-gap format when encoding/encoding RGB.

   // RGB packed in one component. (uint32_t)
   // Will be transformed to YUV_RCT 4:4:4 of appropriate bit-depth.
   LINELET_FMT_X2B10G10R10,
   LINELET_FMT_X8R8G8B8,

   // Planar RGB.
   LINELET_FMT_GBRP8, // 8-bit planar (uint8_t)
   LINELET_FMT_GBRP10, // 10-bit planar (uint16_t)

   // 12-bit
   // Three components in separate planes 4:4:4, 12-bit unsigned, using BT709 color transform. (uint16_t)
   LINELET_FMT_YCBCR12_BT709_444,
   // Three components in separate planes 4:2:2, 12-bit unsigned, using BT709 color transform. (uint16_t)
   LINELET_FMT_YCBCR12_BT709_422,

   // Three components in separate planes 4:4:4, 12-bit unsigned, using BT601 color transform. (uint16_t)
   LINELET_FMT_YCBCR12_BT601_444,
   // Three components in separate planes 4:2:2, 12-bit unsigned, using BT601 color transform. (uint16_t)
   LINELET_FMT_YCBCR12_BT601_422,

   // 10-bit
   // Three components in separate planes 4:4:4, 10-bit signed, reversible color transform. (int16_t)
   LINELET_FMT_YCBCR10_RCT_444,
   // Three components in separate planes 4:2:2, 10-bit signed, reversible color transform. (int16_t)
   LINELET_FMT_YCBCR10_RCT_422,

   // Three components in separate planes 4:4:4, 10-bit unsigned, using BT709 color transform. (uint16_t)
   LINELET_FMT_YCBCR10_BT709_444,
   // Three components in separate planes 4:2:2, 10-bit unsigned, using BT709 color transform. (uint16_t)
   LINELET_FMT_YCBCR10_BT709_422,

   // Three components in separate planes 4:4:4, 10-bit unsigned, using BT601 color transform. (uint16_t)
   LINELET_FMT_YCBCR10_BT601_444,
   // Three components in separate planes 4:2:2, 10-bit unsigned, using BT601 color transform. (uint16_t)
   LINELET_FMT_YCBCR10_BT601_422,

   // 8-bit
   // 8-bit RCT is still packed in 16-bit as a 9-bit range is needed for cb and cr.
   // Three components in separate planes 4:4:4, 8-bit signed, reversible color transform. (int16_t)
   LINELET_FMT_YCBCR8_RCT_444,
   // Three components in separate planes 4:2:2, 8-bit signed, reversible color transform. (int16_t)
   LINELET_FMT_YCBCR8_RCT_422,

   // Three components in separate planes 4:4:4, 8-bit unsigned, using BT709 color transform. (uint8_t)
   LINELET_FMT_YCBCR8_BT709_444,
   // Three components in separate planes 4:2:2, 8-bit unsigned, using BT709 color transform. (uint8_t)
   LINELET_FMT_YCBCR8_BT709_422,

   // Three components in separate planes 4:4:4, 8-bit unsigned, using BT601 color transform. (uint8_t)
   LINELET_FMT_YCBCR8_BT601_444,
   // Three components in separate planes 4:2:2, 8-bit unsigned, using BT601 color transform. (uint8_t)
   LINELET_FMT_YCBCR8_BT601_422,


   LINELET_FMT_DUMMY = INT_MAX
};

unsigned linelet_format_to_subsample(enum linelet_format_format fmt);
unsigned linelet_format_to_bpp(enum linelet_format_format fmt);
enum linelet_format_format linelet_format_parse(const char *arg, const char *matrix);
enum linelet_yuv_transform linelet_format_to_transform(enum linelet_format_format fmt);
unsigned linelet_format_to_components(enum linelet_format_format fmt);
unsigned linelet_format_to_bits(enum linelet_format_format fmt);

// Encoder
typedef struct linelet_format_encoder linelet_format_encoder_t;

struct linelet_format_ratio
{
   uint32_t num;
   uint32_t den;
};

struct linelet_format_encoder_options
{
   enum linelet_format_format format;
   unsigned width;
   unsigned height;
   unsigned precinct_samples;
   unsigned decomposition_levels;
   unsigned slice_height;

   unsigned threads;

   struct linelet_format_ratio sample_aspect_ratio;
   struct linelet_format_ratio frame_rate;
   enum linelet_encoder_rate_allocation_style rate_control;
   unsigned q;

   unsigned filter;
};

linelet_format_encoder_t *linelet_format_encoder_new(const struct linelet_format_encoder_options *opts);

void linelet_format_encoder_free(linelet_format_encoder_t *enc);

// Synchronized frame encoding.
size_t linelet_format_encoder_process_frame(linelet_format_encoder_t *enc, const void * const *components, const size_t *strides,
      void *buffer, size_t buffer_size);

// Asynchronous frame encoding. Allows a main thread to do other things while encoding threads are processing.
bool linelet_format_encoder_process_frame_async(linelet_format_encoder_t *enc, const void * const *components, const size_t *strides,
      void *buffer, size_t buffer_size);
size_t linelet_format_encoder_process_frame_wait(linelet_format_encoder_t *enc);

// Decoder
typedef struct linelet_format_decoder linelet_format_decoder_t;

struct linelet_format_decoder_options
{
   enum linelet_format_format format;
   unsigned threads;
};

#define LINELET_FORMAT_PTS_INVALID ((int64_t)-1)
struct linelet_format_decoder_frame_params
{
   enum linelet_format_format format;
   unsigned width;
   unsigned height;
   struct linelet_format_ratio sample_aspect_ratio;
   struct linelet_format_ratio frame_rate;
   int64_t pts;
};

bool linelet_format_decoder_parse_header(struct linelet_format_decoder_frame_params *params, const void *buffer, size_t buffer_size);

linelet_format_decoder_t *linelet_format_decoder_new(const struct linelet_format_decoder_options *opts);

void linelet_format_decoder_free(linelet_format_decoder_t *dec);

// Synchronized interface.
bool linelet_format_decoder_process_frame(linelet_format_decoder_t *dec,
      const void *encoded_buffer, size_t encoded_buffer_size,
      void * const *components, const size_t *component_strides);

// Async interface.
bool linelet_format_decoder_process_frame_async(linelet_format_decoder_t *dec,
      const void *encoded_buffer, size_t encoded_buffer_size,
      void * const *components, const size_t *component_strides);
bool linelet_format_decoder_process_frame_wait(linelet_format_decoder_t *dec);

#ifdef __cplusplus
}
#endif

#endif

