#include "image.h"
#include "rpng.h"
#include "convert.h"
#include "tiff.h"
#include "memory.h"
#include "log.h"
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <math.h>
#include <zlib.h>

bool linelet_yuv_image_alloc(struct linelet_yuv_image *img, size_t width, size_t height,
      unsigned sub_luma, unsigned sub_cb, unsigned sub_cr)
{
   linelet_yuv_image_free(img);

   img->luma.data = linelet_malloc((width >> sub_luma) * height * sizeof(int16_t));
   img->cb.data = linelet_malloc((width >> sub_cb) * height * sizeof(int16_t));
   img->cr.data = linelet_malloc((width >> sub_cr) * height * sizeof(int16_t));

   img->luma.stride = (width >> sub_luma) * sizeof(int16_t);
   img->cb.stride = (width >> sub_cb) * sizeof(int16_t);
   img->cr.stride = (width >> sub_cr) * sizeof(int16_t);

   img->luma.sub = sub_luma;
   img->cb.sub = sub_cb;
   img->cr.sub = sub_cr;

   img->width = width;
   img->height = height;

   if (!img->luma.data || !img->cb.data || !img->cr.data)
      goto error;

   img->components[0] = img->luma.data;
   img->components[1] = img->cb.data;
   img->components[2] = img->cr.data;
   img->strides[0] = img->luma.stride;
   img->strides[1] = img->cb.stride;
   img->strides[2] = img->cr.stride;

   return true;

error:
   linelet_yuv_image_free(img);
   return false;
}

bool linelet_rgb_image_alloc(struct linelet_rgb_image *img, size_t width, size_t height)
{
   linelet_rgb_image_free(img);

   img->size = width * height * sizeof(uint32_t);
   img->rgb = linelet_malloc(img->size);
   img->stride = width * sizeof(uint32_t);
   img->width = width;
   img->height = height;

   if (!img->rgb)
      goto error;

   return true;

error:
   linelet_rgb_image_free(img);
   return false;
}


static bool linelet_yuv_image_load_png(const char *path, struct linelet_yuv_image *img)
{
   uint32_t *pixels = NULL;
   unsigned width = 0;
   unsigned height = 0;

   rpng_set_malloc(linelet_malloc);
   rpng_set_calloc(linelet_calloc);
   rpng_set_free(linelet_free);
   if (!rpng_load_image_a2bgr10(path, &pixels, &width, &height))
      return false;

   linelet_log_printf(LINELET_LOG_INFO, "Loading image: Width: %u, Height: %u.\n", width, height);

   if (!linelet_yuv_image_alloc(img, width, height, 0, 0, 0))
      goto error;

   linelet_rct_forward_a2bgr10_yuv10(img->components, width * sizeof(int16_t),
         pixels, width, height, width * sizeof(uint32_t));

   img->bits = 10;
   linelet_free(pixels);
   return true;

error:
   linelet_free(pixels);
   linelet_yuv_image_free(img);
   return false;
}

static bool linelet_yuv_image_load_tif(const char *path, struct linelet_yuv_image *img)
{
   uint32_t *pixels = NULL;
   unsigned width = 0;
   unsigned height = 0;

   if (!linelet_tiff_load_image_a2bgr10(path, &pixels, &width, &height))
      return false;

   linelet_log_printf(LINELET_LOG_INFO, "Loading image: Width: %u, Height: %u.\n",
         width, height);

   if (!linelet_yuv_image_alloc(img, width, height, 0, 0, 0))
      goto error;

   linelet_rct_forward_a2bgr10_yuv10(img->components, width * sizeof(int16_t),
         pixels, width, height, width * sizeof(uint32_t));

   img->bits = 10;
   linelet_free(pixels);
   return true;

error:
   linelet_free(pixels);
   linelet_yuv_image_free(img);
   return false;
}

static bool linelet_rgb_image_load_png(const char *path, struct linelet_rgb_image *img)
{
   uint32_t *pixels = NULL;
   unsigned width = 0;
   unsigned height = 0;

   rpng_set_malloc(linelet_malloc);
   rpng_set_calloc(linelet_calloc);
   rpng_set_free(linelet_free);
   if (!rpng_load_image_a2bgr10(path, &pixels, &width, &height))
      return false;

   linelet_log_printf(LINELET_LOG_INFO, "Loading image: Width: %u, Height: %u.\n", width, height);
   img->rgb = pixels;
   img->size = width * height * sizeof(uint32_t);
   img->stride = width * sizeof(uint32_t);
   img->width = width;
   img->height = height;
   img->bits = 10;

   return true;
}

static bool linelet_rgb_image_load_tif(const char *path, struct linelet_rgb_image *img)
{
   uint32_t *pixels = NULL;
   unsigned width = 0;
   unsigned height = 0;

   if (!linelet_tiff_load_image_a2bgr10(path, &pixels, &width, &height))
      return false;

   linelet_log_printf(LINELET_LOG_INFO, "Loading image: Width: %u, Height: %u.\n", width, height);
   img->rgb = pixels;
   img->size = width * height * sizeof(uint32_t);
   img->stride = width * sizeof(uint32_t);
   img->width = width;
   img->height = height;
   img->bits = 10;

   return true;
}

bool linelet_yuv_image_load(const char *path, struct linelet_yuv_image *img)
{
   linelet_yuv_image_free(img);
   const char *ext = strrchr(path, '.');
   if (!ext)
      return false;

   ext++;
   if (!strcmp(ext, "png"))
      return linelet_yuv_image_load_png(path, img);
   else if (!strcmp(ext, "tif"))
      return linelet_yuv_image_load_tif(path, img);
   else
      return false;
}

bool linelet_rgb_image_load(const char *path, struct linelet_rgb_image *img)
{
   linelet_rgb_image_free(img);
   const char *ext = strrchr(path, '.');
   if (!ext)
      return false;

   ext++;
   if (!strcmp(ext, "png"))
      return linelet_rgb_image_load_png(path, img);
   else if (!strcmp(ext, "tif"))
      return linelet_rgb_image_load_tif(path, img);
   else
      return false;
}

bool linelet_yuv_image_save(const char *path, const struct linelet_yuv_image *img)
{
   uint32_t *pixels = linelet_malloc(img->width * img->height * sizeof(uint32_t));
   if (!pixels)
      return false;

   linelet_rct_inverse_yuv10_a2bgr10(pixels, img->width * sizeof(uint32_t),
         (const int16_t**)img->components,
         img->width, img->height, img->luma.stride);

   linelet_log_printf(LINELET_LOG_INFO, "Saving image: Width: %u, Height: %u.\n", (unsigned)img->width, (unsigned)img->height);

   bool ret = rpng_save_image_a2bgr10(path, pixels, img->width, img->height, img->width * sizeof(uint32_t), 1);
   linelet_free(pixels);
   return ret;
}

bool linelet_rgb_image_save(const char *path, const struct linelet_rgb_image *img)
{
   linelet_log_printf(LINELET_LOG_INFO, "Saving image: Width: %u, Height: %u.\n", (unsigned)img->width, (unsigned)img->height);
   return rpng_save_image_a2bgr10(path, img->rgb, img->width, img->height, img->stride, 1);
}

void linelet_yuv_image_free(struct linelet_yuv_image *img)
{
   if (!img)
      return;
   linelet_free(img->luma.data);
   linelet_free(img->cb.data);
   linelet_free(img->cr.data);
   memset(img, 0, sizeof(*img));
}

void linelet_rgb_image_free(struct linelet_rgb_image *img)
{
   if (!img)
      return;
   linelet_free(img->rgb);
   memset(img, 0, sizeof(*img));
}

bool linelet_yuv_image_psnr(struct linelet_yuv_psnr *psnr, const struct linelet_yuv_image *a, const struct linelet_yuv_image *b, unsigned bits)
{
   unsigned height = a->height < b->height ? a->height : b->height;

   memset(psnr, 0, sizeof(*psnr));
   if (a->width != b->width)
      return false;

   const int16_t *a_luma = a->luma.data;
   const int16_t *a_cb = a->cb.data;
   const int16_t *a_cr = a->cr.data;

   const int16_t *b_luma = b->luma.data;
   const int16_t *b_cb = b->cb.data;
   const int16_t *b_cr = b->cr.data;

   size_t a_stride = a->luma.stride >> 1;
   size_t b_stride = b->luma.stride >> 1;

   uint64_t error_luma = 0;
   uint64_t error_cb = 0;
   uint64_t error_cr = 0;

   for (size_t y = 0; y < height; y++, a_luma += a_stride, b_luma += b_stride)
   {
      for (size_t x = 0; x < a->width; x++)
      {
         int16_t diff_luma = a_luma[x] - b_luma[x];
         error_luma += diff_luma * diff_luma;
      }
   }

   for (size_t y = 0; y < height; y++, a_cb += a_stride, b_cb += b_stride)
   {
      for (size_t x = 0; x < a->width; x++)
      {
         int16_t diff_cb = a_cb[x] - b_cb[x];
         error_cb += diff_cb * diff_cb;
      }
   }

   for (size_t y = 0; y < height; y++, a_cr += a_stride, b_cr += b_stride)
   {
      for (size_t x = 0; x < a->width; x++)
      {
         int16_t diff_cr = a_cr[x] - b_cr[x];
         error_cr += diff_cr * diff_cr;
      }
   }

   double mod = 1.0 / (a->width * height);
   psnr->mse.luma = error_luma * mod;
   psnr->mse.cb = error_cb * mod;
   psnr->mse.cr = error_cr * mod;

   double range = (1 << bits) - 1;
   range *= range;
   psnr->luma = 10.0 * log10(range / psnr->mse.luma);
   psnr->cb   = 10.0 * log10(range / psnr->mse.cb);
   psnr->cr   = 10.0 * log10(range / psnr->mse.cr);
   return true;
}

static inline int16_t clamp(int16_t v, int16_t lo, int16_t hi)
{
   if (v < lo)
      return lo;
   else if (v > hi)
      return hi;
   else
      return v;
}

bool linelet_rgb_image_compute_error_image(struct linelet_rgb_image *out, const struct linelet_rgb_image *ref, const struct linelet_rgb_image *res)
{
   if (ref->width != res->width || ref->height != res->height)
      return false;
   if (!linelet_rgb_image_alloc(out, ref->width, ref->height))
      return false;

   const uint32_t *ref_rgb = ref->rgb;
   const uint32_t *res_rgb = res->rgb;
   uint32_t *out_rgb = out->rgb;

   size_t ref_stride = ref->stride >> 2;
   size_t res_stride = res->stride >> 2;
   size_t out_stride = out->stride >> 2;

   for (size_t y = 0; y < ref->height; y++, ref_rgb += ref_stride, res_rgb += res_stride, out_rgb += out_stride)
   {
      for (size_t x = 0; x < ref->width; x++)
      {
         uint32_t a = ref_rgb[x];
         uint32_t b = res_rgb[x];
         int16_t diff_r = ((a >>  0) & 0x3ff) - ((b >>  0) & 0x3ff);
         int16_t diff_g = ((a >> 10) & 0x3ff) - ((b >> 10) & 0x3ff);
         int16_t diff_b = ((a >> 20) & 0x3ff) - ((b >> 20) & 0x3ff);
         diff_r <<= 5;
         diff_g <<= 5;
         diff_b <<= 5;
         diff_r = clamp(diff_r, -512, 511);
         diff_g = clamp(diff_g, -512, 511);
         diff_b = clamp(diff_b, -512, 511);
         diff_r += 512;
         diff_g += 512;
         diff_b += 512;
         out_rgb[x] = (0x3u << 30) | (diff_b << 20) | (diff_g << 10) | (diff_r << 0);
      }
   }

   return true;
}

bool linelet_rgb_image_psnr(struct linelet_rgb_psnr *psnr, const struct linelet_rgb_image *a, const struct linelet_rgb_image *b, unsigned bits)
{
   unsigned height = a->height < b->height ? a->height : b->height;
   memset(psnr, 0, sizeof(*psnr));

   if (a->width != b->width)
      return false;

   const uint32_t *a_rgb = a->rgb;
   const uint32_t *b_rgb = b->rgb;

   size_t a_stride = a->stride >> 2;
   size_t b_stride = b->stride >> 2;

   uint64_t error_r = 0;
   uint64_t error_g = 0;
   uint64_t error_b = 0;

   for (size_t y = 0; y < height; y++, a_rgb += a_stride, b_rgb += b_stride)
   {
      for (size_t x = 0; x < a->width; x++)
      {
         uint32_t a_col = a_rgb[x];
         uint32_t b_col = b_rgb[x];

         uint16_t a_r = (a_col >>  0) & 0x3ff;
         uint16_t a_g = (a_col >> 10) & 0x3ff;
         uint16_t a_b = (a_col >> 20) & 0x3ff;

         uint16_t b_r = (b_col >>  0) & 0x3ff;
         uint16_t b_g = (b_col >> 10) & 0x3ff;
         uint16_t b_b = (b_col >> 20) & 0x3ff;

         int16_t diff_r = a_r - b_r;
         int16_t diff_g = a_g - b_g;
         int16_t diff_b = a_b - b_b;
         error_r += diff_r * diff_r;
         error_g += diff_g * diff_g;
         error_b += diff_b * diff_b;
      }
   }

   double mod = 1.0 / (a->width * height);
   psnr->mse.r = error_r * mod;
   psnr->mse.g = error_g * mod;
   psnr->mse.b = error_b * mod;

   double range = (1 << bits) - 1;
   range *= range;
   psnr->r = 10.0 * log10(range / psnr->mse.r);
   psnr->g = 10.0 * log10(range / psnr->mse.g);
   psnr->b = 10.0 * log10(range / psnr->mse.b);
   return true;
}

double linelet_image_component_mse_8(const uint8_t *a, const uint8_t *b, unsigned samples)
{
   double error = 0.0;
   for (unsigned i = 0; i < samples; i++)
      error += (a[i] - b[i]) * (a[i] - b[i]);
   return error / samples;
}

double linelet_image_component_mse_16(const uint16_t *a, const uint16_t *b, unsigned samples)
{
   double error = 0.0;
   for (unsigned i = 0; i < samples; i++)
      error += (a[i] - b[i]) * (a[i] - b[i]);
   return error / samples;
}

static double mean16(const uint16_t *a, unsigned width, unsigned height, unsigned step)
{
   double sum = 0.0;
   for (unsigned y = 0; y < height; y++, a += step)
      for (unsigned x = 0; x < width; x++)
         sum += a[x];
   return sum / (width * height);
}

static double corr16(const uint16_t *a, const uint16_t *b,
      unsigned width, unsigned height, unsigned step)
{
   double sum = 0.0;
   for (unsigned y = 0; y < height; y++, a += step, b += step)
      for (unsigned x = 0; x < width; x++)
         sum += a[x] * b[x];
   return sum / (width * height);
}

static double mean8(const uint8_t *a, unsigned width, unsigned height, unsigned step)
{
   double sum = 0.0;
   for (unsigned y = 0; y < height; y++, a += step)
      for (unsigned x = 0; x < width; x++)
         sum += a[x];
   return sum / (width * height);
}

static double corr8(const uint8_t *a, const uint8_t *b,
      unsigned width, unsigned height, unsigned step)
{
   double sum = 0.0;
   for (unsigned y = 0; y < height; y++, a += step, b += step)
      for (unsigned x = 0; x < width; x++)
         sum += a[x] * b[x];
   return sum / (width * height);
}

static double ssim_block_16(const uint16_t *x, const uint16_t *y,
      unsigned width, double c1, double c2)
{
   double mean_x = mean16(x, 8, 8, width);
   double mean_y = mean16(y, 8, 8, width);
   double var_x = corr16(x, x, 8, 8, width) - mean_x * mean_x;
   double var_y = corr16(y, y, 8, 8, width) - mean_y * mean_y;
   double covar = corr16(x, y, 8, 8, width) - mean_x * mean_y;

   return (2.0 * mean_x * mean_y + c1) * (2.0 * covar + c2) /
      ((mean_x * mean_x + mean_y * mean_y + c1) * (var_x + var_y + c2));
}

static double ssim_block_8(const uint8_t *x, const uint8_t *y,
      unsigned width, double c1, double c2)
{
   double mean_x = mean8(x, 8, 8, width);
   double mean_y = mean8(y, 8, 8, width);
   double var_x = corr8(x, x, 8, 8, width) - mean_x * mean_x;
   double var_y = corr8(y, y, 8, 8, width) - mean_y * mean_y;
   double covar = corr8(x, y, 8, 8, width) - mean_x * mean_y;

   return (2.0 * mean_x * mean_y + c1) * (2.0 * covar + c2) /
      ((mean_x * mean_x + mean_y * mean_y + c1) * (var_x + var_y + c2));
}

double linelet_image_component_ssim_16(const uint16_t *a, const uint16_t *b,
      unsigned width, unsigned height, unsigned bits)
{
   double c1 = 0.01 * ((1 << bits) - 1);
   double c2 = 0.03 * ((1 << bits) - 1);
   c1 *= c1;
   c2 *= c2;

   double ssim_total = 0.0;
   unsigned blocks = 0;
   for (unsigned y = 0; y + 8 <= height; y += 4)
      for (unsigned x = 0; x + 8 <= width; x += 4, blocks++)
         ssim_total += ssim_block_16(a + y * width + x, b + y * width + x, width, c1, c2);

   return ssim_total / blocks;
}

double linelet_image_component_ssim_8(const uint8_t *a, const uint8_t *b,
      unsigned width, unsigned height)
{
   double c1 = 0.01 * 255.0;
   double c2 = 0.03 * 255.0;
   c1 *= c1;
   c2 *= c2;

   double ssim_total = 0.0;
   unsigned blocks = 0;
   for (unsigned y = 0; y + 8 <= height; y += 4)
      for (unsigned x = 0; x + 8 <= width; x += 4, blocks++)
         ssim_total += ssim_block_8(a + y * width + x, b + y * width + x, width, c1, c2);

   return ssim_total / blocks;
}
