#include "linelet.h"
#include "pack.h"
#include "quantizer.h"
#include "memory.h"
#include "legall.h"
#include "haar.h"
#include "util.h"
#include "log.h"
#include <stdlib.h>

#define LINELET_VERTICAL_HAAR 0

struct linelet_decoder
{
   struct linelet_wavelet_buffer buffer[LINELET_MAX_COMPONENTS];

   unsigned samples;
   unsigned precinct_samples;
   unsigned precincts;
   unsigned lines;
   unsigned lines_log2;

   uint8_t *bits;
   
   struct linelet_component components[LINELET_MAX_COMPONENTS];
   unsigned num_components;
   unsigned levels[LINELET_MAX_COMPONENTS][LINELET_MAX_LINES];
};

void linelet_decoder_free(linelet_decoder_t *ll)
{
   if (!ll)
      return;

   linelet_free(ll->bits);
   linelet_common_free_buffers(ll->buffer, LINELET_MAX_COMPONENTS);
   linelet_free(ll);
}

unsigned linelet_decoder_get_precinct_samples(const linelet_decoder_t *dec)
{
   return dec->precinct_samples;
}

linelet_decoder_t *linelet_decoder_new(unsigned samples,
      unsigned lines,
      const struct linelet_component *components, unsigned num_components,
      const struct linelet_decoder_options *opts)
{
   linelet_init();
   if (!lines || (lines & (lines - 1)))
   {
      linelet_log_printf(LINELET_LOG_ERROR, "[Decoder]: Lines must be positive and power-of-two.\n");
      return NULL;
   }

   if (lines > LINELET_MAX_LINES)
   {
      linelet_log_printf(LINELET_LOG_ERROR, "[Decoder]: Max lines is 8.\n");
      return NULL;
   }

   // POW2 check.
   if (opts->precinct_samples & (opts->precinct_samples - 1))
      return NULL;

   linelet_decoder_t *ll = linelet_calloc(1, sizeof(*ll));
   if (!ll)
      return NULL;

   // Divisible checks.
   if (samples & (opts->precinct_samples - 1))
      goto error;

   unsigned minimum_decomp = linelet_log2i(lines);
   for (unsigned i = 0; i < num_components; i++)
   {
      unsigned samples_mask = (1 << components[i].decomposition_levels) - 1;
      if ((samples >> components[i].decimation) & samples_mask)
      {
         linelet_log_printf(LINELET_LOG_ERROR, "[Decoder]: Vertical samples not divisible by 2^decompositions.\n");
         goto error;
      }

      if (components[i].decomposition_levels < minimum_decomp)
      {
         linelet_log_printf(LINELET_LOG_ERROR, "[Decoder]: Horizontal decomposition levels less than vertical decompositions.\n");
         goto error;
      }

      if ((opts->precinct_samples >> components[i].decomposition_levels) < 2)
      {
         linelet_log_printf(LINELET_LOG_ERROR, "[Decoder]: Precinct samples too low.\n");
         goto error;
      }

      ll->components[i] = components[i];
      if (!components[i].decomposition_levels || components[i].decomposition_levels > LINELET_MAX_DECOMP_LEVELS)
      {
         linelet_log_printf(LINELET_LOG_ERROR, "[Decoder]: Decompositions invalid.\n");
         goto error;
      }

      linelet_common_set_component_levels(ll->levels[i], lines, components[i].decomposition_levels);
   }

   ll->samples = samples;
   ll->lines = lines;
   ll->lines_log2 = linelet_log2i(lines);
   ll->precinct_samples = opts->precinct_samples;
   ll->precincts = ll->samples / ll->precinct_samples;
   ll->num_components = num_components;
   if (!(ll->bits = linelet_malloc(ll->precincts)))
      goto error;

   if (!linelet_common_allocate_buffers(ll->buffer, ll->samples, ll->lines, ll->components, ll->num_components))
      goto error;

   return ll;

error:
   linelet_decoder_free(ll);
   return NULL;
}

static size_t linelet_decoder_process_component_line(linelet_decoder_t *ll,
      unsigned c, unsigned line, const uint8_t *buffer)
{
   const uint8_t *orig_buffer = buffer;
   const struct linelet_wavelet_buffer *wave = &ll->buffer[c];

   unsigned levels = ll->levels[c][line];

   // Unpack quantization params.
   uint8_t quants[LINELET_MAX_BANDS];
   buffer += linelet_unpack_nibbles_u8(quants, buffer, levels + 1); 

   unsigned precinct_step = ll->precinct_samples >> ll->components[c].decimation;
   unsigned samples = ll->samples >> ll->components[c].decimation;

   for (unsigned b = 0; b <= levels; b++)
   {
      unsigned shift = b == levels ? levels : (b + 1);
      int16_t *band = b == levels ? wave->lp[line][levels - 1] : wave->hp[line][b];

      buffer += linelet_unpack_predictive_unary(ll->bits, buffer, ll->precincts);
      unsigned step = precinct_step >> shift;

      buffer += linelet_unpack_bits_s16(band, buffer, step, ll->precincts, ll->bits);
      linelet_dequantize_samples(band, samples >> shift, quants[b]);
   }

   return buffer - orig_buffer;
}

static void linelet_dwt_inverse(linelet_decoder_t *ll,
      int16_t *output, size_t component_stride, unsigned c)
{
   const struct linelet_wavelet_buffer *buffer = &ll->buffer[c];
   unsigned lines = ll->lines;
   unsigned l = ll->levels[c][0] - 1;
   unsigned samples = ll->samples >> (ll->components[c].decimation + l);
   component_stride >>= 1;

   // Reconstruct LL until we start doing inverse vertical as well.
   unsigned levels_2d = ll->lines_log2;
   for (; l >= levels_2d && l; l--, samples <<= 1)
      linelet_legall_dwt_inverse(buffer->lp[0][l - 1],
            buffer->lp[0][l], buffer->hp[0][l], samples);

   for (unsigned step = lines >> 1; step > 1; step >>= 1, samples <<= 1, l--)
   {
      for (unsigned y = 0; y < lines; y += step << 1)
      {
         // Inverse horizontal.
         linelet_legall_dwt_inverse(buffer->lp[y][l - 1],
               buffer->lp[y][l], buffer->hp[y][l], samples);
         linelet_legall_dwt_inverse(buffer->lp[y + step][l - 1],
               buffer->lp[y + step][l], buffer->hp[y + step][l], samples);
      }

#if LINELET_VERTICAL_HAAR
      for (unsigned y = 0; y < lines; y += step << 1)
      {
         int16_t *a = buffer->lp[y][l - 1];
         int16_t *b = buffer->lp[y + step][l - 1];
         linelet_haar_dwt_inverse(a, b, samples);
      }
#else
      unsigned vert_lines = lines >> l;
      if (vert_lines > 2)
      {
         linelet_legall_vert_dwt_inverse_top(
               buffer->lp[0 << l][l - 1],
               buffer->lp[1 << l][l - 1],
               samples);

         for (unsigned y = 2; y + 2 <= vert_lines; y += 2)
         {
            linelet_legall_vert_dwt_inverse(
               buffer->lp[(y - 2) << l][l - 1],
               buffer->lp[(y - 1) << l][l - 1],
               buffer->lp[(y + 0) << l][l - 1],
               buffer->lp[(y + 1) << l][l - 1],
               samples);
         }

         linelet_legall_vert_dwt_inverse_bottom(
               buffer->lp[(vert_lines - 2) << l][l - 1],
               buffer->lp[(vert_lines - 1) << l][l - 1],
               samples);
      }
      else
      {
         // Inverse lift vertically.
         int16_t *a = buffer->lp[0][l - 1];
         int16_t *b = buffer->lp[step][l - 1];
         linelet_haar_dwt_inverse(a, b, samples);
      }
#endif
   }

   // Now we have LL, LH, HL, HH. Finish it off.
   for (unsigned y = 0; y < lines; y++)
      linelet_legall_dwt_inverse(output + component_stride * y, buffer->lp[y][0], buffer->hp[y][0], samples);

#if LINELET_VERTICAL_HAAR
   if (lines > 1)
   {
      int16_t *dst = output;
      for (unsigned y = 0; y < lines; y += 2, dst += component_stride << 1)
         linelet_haar_dwt_inverse(dst, dst + component_stride, samples);
   }
#else
   // If no vertical transform, we only had L, H.
   if (lines > 2)
   {
      linelet_legall_vert_dwt_inverse_top(
            output + 0 * component_stride,
            output + 1 * component_stride,
            samples);

      for (unsigned y = 2; y + 2 <= lines; y += 2)
      {
         linelet_legall_vert_dwt_inverse(
               output + (y - 2) * component_stride,
               output + (y - 1) * component_stride,
               output + (y + 0) * component_stride,
               output + (y + 1) * component_stride,
               samples);
      }

      linelet_legall_vert_dwt_inverse_bottom(
            output + (lines - 2) * component_stride,
            output + (lines - 1) * component_stride,
            samples);
   }
   else if (lines == 2) // If lines == 2, Haar is equivalent to Legall.
      linelet_haar_dwt_inverse(output, output + component_stride, samples);
#endif
}

size_t linelet_decoder_process(linelet_decoder_t *ll,
      const void *encoder_buffer_,
      int16_t * const *component_buffers, const size_t *component_strides)
{
   size_t decoded_size = 0;
   const uint8_t *encoder_buffer = encoder_buffer_;

   for (unsigned c = 0; c < ll->num_components; c++)
      for (unsigned l = 0; l < ll->lines; l++)
         decoded_size += linelet_decoder_process_component_line(ll, c, l, encoder_buffer + decoded_size);

   for (unsigned c = 0; c < ll->num_components; c++)
      linelet_dwt_inverse(ll, component_buffers[c], component_strides[c], c);

   return decoded_size;
}

