#include "linelet.h"
#include "rate_allocator.h"
#include "quantizer.h"
#include "pack.h"
#include "legall.h"
#include "memory.h"
#include "haar.h"
#include "util.h"
#include "log.h"
#include <stdbool.h>
#include <stdlib.h>
#include <math.h>

#define GOLOMB_STAT 0
#define BAND_POWER 0
#define BAND_MINMAX 0

#define LINELET_VERTICAL_HAAR 0

struct linelet_encoder
{
   struct linelet_wavelet_buffer buffer[LINELET_MAX_COMPONENTS];

   struct linelet_component components[LINELET_MAX_COMPONENTS];
   unsigned num_components;
   unsigned levels[LINELET_MAX_COMPONENTS][LINELET_MAX_LINES];

   unsigned samples;
   unsigned lines;
   unsigned precinct_samples;
   unsigned precincts;

   uint8_t *bits;
   
   linelet_rate_allocator_t *rate;
   size_t line_overhead;

#if GOLOMB_STAT
   size_t pack_bits;
   size_t optim_pack_bits;
   size_t golomb_bits;
   unsigned distribution[LINELET_MAX_COMPONENTS][LINELET_MAX_LINES][LINELET_MAX_BANDS][4096];
   unsigned universal_distribution[4096];

   unsigned precinct_bits_pred_distribution[2 * 16 + 1];
   unsigned precinct_bits_distribution[16 + 1];
   size_t precinct_bits_pred_unary;
   size_t precinct_bits_unary;
#endif

#if BAND_POWER
   uint64_t powers_total;
   uint64_t powers_total_pre;
   uint64_t powers_total_count;
   uint64_t powers[LINELET_MAX_COMPONENTS][LINELET_MAX_LINES][LINELET_MAX_BANDS];
   uint64_t power_counts[LINELET_MAX_COMPONENTS][LINELET_MAX_LINES][LINELET_MAX_BANDS];
#endif

#if BAND_MINMAX
   int16_t band_min[LINELET_MAX_COMPONENTS][LINELET_MAX_LINES][LINELET_MAX_BANDS];
   int16_t band_max[LINELET_MAX_COMPONENTS][LINELET_MAX_LINES][LINELET_MAX_BANDS];
#endif
};

#if GOLOMB_STAT
size_t compute_entropy(const unsigned *counts, unsigned num_counts)
{
   double entropy = 0.0;
   size_t total = 0;
   for (unsigned i = 0; i < num_counts; i++)
      total += counts[i];
   for (unsigned i = 0; i < num_counts; i++)
   {
      if (counts[i])
      {
         double p = (double)counts[i] / total;
         entropy += -p * log2(p);
      }
   }

   return entropy * total;
}
#endif

void linelet_encoder_free(linelet_encoder_t *ll)
{
   if (!ll)
      return;

#if GOLOMB_STAT
   size_t split_entropy = 0;
   size_t entropy;
   linelet_log_printf(LINELET_LOG_DEBUG, "[Distribution]:\n");
   linelet_log_printf(LINELET_LOG_DEBUG, "%25s: %10zu bits\n", "Pack", ll->pack_bits);
   linelet_log_printf(LINELET_LOG_DEBUG, "%25s: %10zu bits\n", "Exp-golomb", ll->golomb_bits);

   linelet_log_printf(LINELET_LOG_DEBUG, "%25s: %10zu bits\n", "Optim-Pack", ll->optim_pack_bits);
   entropy = compute_entropy(ll->precinct_bits_pred_distribution, 2 * 16 + 1);
   linelet_log_printf(LINELET_LOG_DEBUG, "%25s: %10zu bits\n", "Precinct-pred", entropy);
   entropy = compute_entropy(ll->precinct_bits_distribution, 16 + 1);
   linelet_log_printf(LINELET_LOG_DEBUG, "%25s: %10zu bits\n", "Precinct", entropy);
   linelet_log_printf(LINELET_LOG_DEBUG, "%25s: %10zu bits\n", "Precinct-pred-unary", ll->precinct_bits_pred_unary);
   linelet_log_printf(LINELET_LOG_DEBUG, "%25s: %10zu bits\n", "Precinct-unary", ll->precinct_bits_unary);

   entropy = compute_entropy(ll->universal_distribution, 4096);
   linelet_log_printf(LINELET_LOG_DEBUG, "%25s: %10zu bits\n", "Universal", entropy);
   for (unsigned c = 0; c < ll->num_components; c++)
   {
      for (unsigned line = 0; line < ll->lines; line++)
      {
         for (unsigned b = 0; b <= ll->levels[c][line]; b++)
         {
            entropy = compute_entropy(ll->distribution[c][line][b], 4096);
            split_entropy += entropy;
         }
      }
   }
   linelet_log_printf(LINELET_LOG_DEBUG, "%25s: %10zu bits\n", "Split", split_entropy);
#endif

#if BAND_POWER
   for (unsigned c = 0; c < 1 /*ll->num_components*/; c++)
      for (unsigned line = 0; line < ll->lines; line++)
         for (unsigned b = 0; b <= ll->levels[c][line]; b++)
            linelet_log_printf(LINELET_LOG_DEBUG, "[Power]: C = %u, L = %u, B = %u, Mean: %.3f\n", c, line, b,
                  (double)ll->powers[c][line][b] / ll->power_counts[c][line][b]);
   linelet_log_printf(LINELET_LOG_DEBUG, "[Power]: Overall (post transform): %.3f\n", (double)ll->powers_total / ll->powers_total_count);
   linelet_log_printf(LINELET_LOG_DEBUG, "[Power]: Overall (pre transform): %.3f\n", (double)ll->powers_total_pre / ll->powers_total_count);
#endif

#if BAND_MINMAX
   for (unsigned c = 0; c < ll->num_components; c++)
      for (unsigned line = 0; line < ll->lines; line++)
         for (unsigned b = 0; b <= ll->levels[c][line]; b++)
            linelet_log_printf(LINELET_LOG_DEBUG, "C = %u, L = %u, B = %u, (%d, %d)\n",
                  c, line, b, ll->band_min[c][line][b], ll->band_max[c][line][b]);
#endif

   linelet_free(ll->bits);
   linelet_rate_allocator_free(ll->rate);
   linelet_common_free_buffers(ll->buffer, LINELET_MAX_COMPONENTS);

   linelet_free(ll);
}

static size_t linelet_encoder_compute_overhead(linelet_encoder_t *ll)
{
   size_t line_overhead = 0;

   for (unsigned line = 0; line < ll->lines; line++)
   {
      for (unsigned c = 0; c < ll->num_components; c++)
      {
         // Quantization parameters at start of scanline.
         // 4-bit Q factors for each band packed. Each component's coeffs aligned on byte boundary.
         unsigned levels = ll->levels[c][line];
         line_overhead += (((levels + 1) << 2) + 7) >> 3;
      }
   }

   return line_overhead;
}

linelet_encoder_t *linelet_encoder_new(unsigned samples,
      unsigned lines,
      const struct linelet_component *components, unsigned num_components,
      const struct linelet_encoder_options *opts)
{
   linelet_init();
   if (!lines || (lines & (lines - 1)))
   {
      linelet_log_printf(LINELET_LOG_ERROR, "[Encoder]: Lines must be positive and power-of-two.\n");
      return NULL;
   }

   if (lines > LINELET_MAX_LINES)
   {
      linelet_log_printf(LINELET_LOG_ERROR, "[Encoder]: Max lines is 8.\n");
      return NULL;
   }

   // POW2 check.
   if (opts->precinct_samples & (opts->precinct_samples - 1))
      return NULL;

   linelet_encoder_t *ll = linelet_calloc(1, sizeof(*ll));
   if (!ll)
      return NULL;

   // Divisible checks.
   if (samples & (opts->precinct_samples - 1))
      goto error;

   unsigned minimum_decomp = linelet_log2i(lines);
   for (unsigned i = 0; i < num_components; i++)
   {
      unsigned samples_mask = (1 << components[i].decomposition_levels) - 1;
      if ((samples >> components[i].decimation) & samples_mask)
      {
         linelet_log_printf(LINELET_LOG_ERROR, "[Encoder]: Vertical samples not divisible by 2^decompositions.\n");
         goto error;
      }

      if (components[i].decomposition_levels < minimum_decomp)
      {
         linelet_log_printf(LINELET_LOG_ERROR, "[Encoder]: Horizontal decomposition levels less than vertical decompositions.\n");
         goto error;
      }

      if ((opts->precinct_samples >> components[i].decomposition_levels) < 2)
      {
         linelet_log_printf(LINELET_LOG_ERROR, "[Encoder]: Precinct samples too low.\n");
         goto error;
      }

      ll->components[i] = components[i];
      if (!components[i].decomposition_levels || components[i].decomposition_levels > LINELET_MAX_DECOMP_LEVELS)
      {
         linelet_log_printf(LINELET_LOG_ERROR, "[Encoder]: Decompositions invalid.\n");
         goto error;
      }

      linelet_common_set_component_levels(ll->levels[i], lines, components[i].decomposition_levels);
   }

   ll->lines = lines;
   ll->samples = samples;
   ll->precinct_samples = opts->precinct_samples;
   ll->precincts = ll->samples / ll->precinct_samples;
   ll->num_components = num_components;
   if (!(ll->bits = linelet_malloc(ll->precincts)))
      goto error;

   ll->rate = linelet_rate_allocator_new(ll->precinct_samples, ll->samples, ll->lines,
         ll->components, ll->num_components, NULL, 0);
   if (!ll->rate)
      goto error;

   if (opts->rate_control == LINELET_ENCODER_RATE_ALLOCATION_MINIMUM_QUANT)
      linelet_rate_allocator_minimum_quant(ll->rate, opts->q);

   if (!linelet_common_allocate_buffers(ll->buffer, ll->samples, ll->lines, ll->components, ll->num_components))
      goto error;

   ll->line_overhead = linelet_encoder_compute_overhead(ll);
   return ll;

error:
   linelet_encoder_free(ll);
   return NULL;
}

#if GOLOMB_STAT
static unsigned study_golomb_bits(unsigned *dist_a, unsigned *dist_b, const int16_t *data, unsigned samples)
{
   unsigned bits = 0;
   for (unsigned i = 0; i < samples; i++)
   {
      unsigned v = abs(data[i]);
      dist_a[data[i] + 2048]++;
      dist_b[data[i] + 2048]++;

      if (v)
      {
         v++;
         unsigned shift = sizeof(unsigned) * CHAR_BIT - linelet_clz(v) - 1;
         bits += shift * 2 + 2;
      }
      else
         bits++;
   }

   if (bits == samples) // Skip precinct.
      bits = 0;
   return (bits + 7) & ~7;
}
#endif

#if BAND_POWER
uint64_t compute_power(const int16_t *data, unsigned samples)
{
   uint64_t res = 0;
   for (unsigned i = 0; i < samples; i++)
      res += data[i] * data[i];
   return res;
}
#endif

static size_t linelet_encoder_pack_line_component(linelet_encoder_t *ll, uint8_t *buffer, unsigned c, unsigned l)
{
   uint8_t *orig_buffer = buffer;

   unsigned levels = ll->levels[c][l];

   const struct linelet_wavelet_buffer *wave = &ll->buffer[c];

   // Encode quantization params.
   uint8_t quants[LINELET_MAX_BANDS];
   for (unsigned b = 0; b <= levels; b++)
      quants[b] = linelet_rate_allocator_quant_factor(ll->rate, c, l, b);
   buffer += linelet_pack_nibbles_u8(buffer, quants, levels + 1);

   unsigned precinct_step = ll->precinct_samples >> ll->components[c].decimation;

   for (unsigned b = 0; b <= levels; b++)
   {
      unsigned shift = b == levels ? levels : (b + 1);
      int16_t *band = b == levels ? wave->lp[l][levels - 1] : wave->hp[l][b];

      const struct linelet_precinct_range_level *range = linelet_rate_allocator_precincts(ll->rate, c, l, b);
      for (unsigned p = 0; p < ll->precincts; p++)
         ll->bits[p] = range[p].bits;

      buffer += linelet_pack_predictive_unary(buffer, ll->bits, ll->precincts);

      unsigned step = precinct_step >> shift;
      linelet_quantize_samples(band, step * ll->precincts,
            linelet_rate_allocator_quant_factor(ll->rate, c, l, b));

      buffer += linelet_pack_bits_s16(buffer, band, step, ll->precincts, ll->bits);

#if GOLOMB_STAT
      ll->pack_bits += ll->precincts * 4;
      ll->golomb_bits += ll->precincts;
#endif

#if BAND_POWER
      uint64_t power = compute_power(band, step * ll->precincts);
      ll->powers[c][l][b] += power;
      ll->power_counts[c][l][b] += step * ll->precincts;
      ll->powers_total += power;
      ll->powers_total_count += step * ll->precincts;
#endif

#if BAND_MINMAX
      for (unsigned i = 0; i < step * ll->precincts; i++)
      {
         if (band[i] < ll->band_min[c][l][b])
            ll->band_min[c][l][b] = band[i];
         if (band[i] > ll->band_max[c][l][b])
            ll->band_max[c][l][b] = band[i];
      }
#endif

#if GOLOMB_STAT
      for (unsigned p = 0; p < ll->precincts; p++, band += step)
      {
         ll->pack_bits += (ll->bits[p] * step + 7) & ~7;
         ll->optim_pack_bits += (ll->bits[p] * step + 7) & ~7;
         ll->golomb_bits += study_golomb_bits(ll->universal_distribution, ll->distribution[c][l][b], band, step);

         int pred_bits = p ? (ll->bits[p] - ll->bits[p - 1]) : ll->bits[p];
         ll->precinct_bits_pred_distribution[pred_bits + 16]++;
         ll->precinct_bits_distribution[ll->bits[p]]++;
         ll->precinct_bits_pred_unary += pred_bits ? abs(pred_bits) + (p ? !!ll->bits[p - 1] : 0) + 1 : 1;
         ll->precinct_bits_unary += ll->bits[p] + 1;

      }
      ll->precinct_bits_pred_unary = (ll->precinct_bits_pred_unary + 7) & ~7;
      ll->precinct_bits_unary = (ll->precinct_bits_unary + 7) & ~7;
#endif
   }

   return buffer - orig_buffer;
}

static size_t linelet_encoder_pack_lines(linelet_encoder_t *ll, uint8_t *buffer)
{
   size_t scanline_size = 0;

   // All components are packed independently after each other.
   for (unsigned c = 0; c < ll->num_components; c++)
      for (unsigned l = 0; l < ll->lines; l++)
         scanline_size += linelet_encoder_pack_line_component(ll, buffer + scanline_size, c, l);

   return scanline_size;
}

static void linelet_dwt_forward(linelet_encoder_t *ll,
      int16_t *input, size_t component_stride, unsigned c)
{
   const struct linelet_wavelet_buffer *buffer = &ll->buffer[c];
   unsigned lines = ll->lines;
   unsigned samples = ll->samples >> ll->components[c].decimation;
   unsigned levels = ll->levels[c][0];
   component_stride >>= 1;

#if LINELET_VERTICAL_HAAR
   if (lines > 1)
   {
      int16_t *src = input;
      for (unsigned y = 0; y < lines; y += 2, src += component_stride << 1)
         linelet_haar_dwt_forward(src, src + component_stride, samples);
   }
#else
   if (lines > 2)
   {
      linelet_legall_vert_dwt_forward_top(
            input + 0 * component_stride,
            input + 1 * component_stride,
            input + 2 * component_stride,
            samples);

      for (unsigned y = 2; y + 2 < lines; y += 2)
      {
         linelet_legall_vert_dwt_forward(
               input + (y - 1) * component_stride,
               input + (y + 0) * component_stride,
               input + (y + 1) * component_stride,
               input + (y + 2) * component_stride,
               samples);
      }

      linelet_legall_vert_dwt_forward_bottom(
            input + (lines - 3) * component_stride,
            input + (lines - 2) * component_stride,
            input + (lines - 1) * component_stride,
            samples);
   }
   else if (lines == 2) // If lines == 2, Legall simply reduces to Haar.
      linelet_haar_dwt_forward(input, input + component_stride, samples);
#endif

   int16_t *src = input;
   for (unsigned y = 0; y < lines; y++, src += component_stride)
      linelet_legall_dwt_forward(buffer->lp[y][0], buffer->hp[y][0], src, samples);

   // First LL, LH, HL, HH done. Only keep processing the LL band.
   unsigned l = 1;
   samples >>= 1;

   for (unsigned step = 2; step < lines; step <<= 1, samples >>= 1, l++)
   {
#if LINELET_VERTICAL_HAAR
      for (unsigned y = 0; y < lines; y += step << 1)
      {
         int16_t *a = buffer->lp[y][l - 1];
         int16_t *b = buffer->lp[y + step][l - 1];
         linelet_haar_dwt_forward(a, b, samples);
      }
#else
      unsigned vert_lines = lines >> l;

      // Lift LL vertically.
      if (vert_lines > 2)
      {
         linelet_legall_vert_dwt_forward_top(
               buffer->lp[0 << l][l - 1],
               buffer->lp[1 << l][l - 1],
               buffer->lp[2 << l][l - 1],
               samples);

         for (unsigned y = 2; y + 2 < vert_lines; y += 2)
         {
            linelet_legall_vert_dwt_forward(
               buffer->lp[(y - 1) << l][l - 1],
               buffer->lp[(y + 0) << l][l - 1],
               buffer->lp[(y + 1) << l][l - 1],
               buffer->lp[(y + 2) << l][l - 1],
                  samples);
         }

         linelet_legall_vert_dwt_forward_bottom(
               buffer->lp[(vert_lines - 3) << l][l - 1],
               buffer->lp[(vert_lines - 2) << l][l - 1],
               buffer->lp[(vert_lines - 1) << l][l - 1],
               samples);
      }
      else if (vert_lines == 2) // If lines == 2, Legall simply reduces to Haar.
      {
         int16_t *a = buffer->lp[0][l - 1];
         int16_t *b = buffer->lp[step][l - 1];
         linelet_haar_dwt_forward(a, b, samples);
      }
#endif

      // Finish off horizontally.
      for (unsigned y = 0; y < lines; y += step << 1)
      {
         linelet_legall_dwt_forward(buffer->lp[y][l],
               buffer->hp[y][l], buffer->lp[y][l - 1], samples);
         linelet_legall_dwt_forward(buffer->lp[y + step][l],
               buffer->hp[y + step][l], buffer->lp[y + step][l - 1], samples);
      }
   }

   // Finish up the last horizontal transforms on LL if any left.
   for (; l < levels; l++, samples >>= 1)
      linelet_legall_dwt_forward(buffer->lp[0][l], buffer->hp[0][l], buffer->lp[0][l - 1], samples);
}

size_t linelet_encoder_process(linelet_encoder_t *ll,
      void *encoded_buffer, size_t encoded_buffer_size,
      int16_t * const *component_buffers, const size_t *component_strides)
{
   // No way we can pack this.
   if (ll->line_overhead >= encoded_buffer_size)
      return 0;

#if BAND_POWER
   for (unsigned c = 0; c < ll->num_components; c++)
      for (unsigned l = 0; l < ll->lines; l++)
         ll->powers_total_pre += compute_power(component_buffers[c] + (component_strides[c] >> 1) * l, ll->samples >> ll->components[c].decimation);
#endif

   // Forward DWT transforms.
   for (unsigned c = 0; c < ll->num_components; c++)
      linelet_dwt_forward(ll, component_buffers[c], component_strides[c], c);

   // Rate allocation. Find quantization factors which allow us to pack all coeffs.
   // Quantization params are separate for each line and band (but not precinct).
   linelet_rate_allocator_study_lines(ll->rate, ll->buffer, encoded_buffer_size - ll->line_overhead);

   size_t ret = linelet_encoder_pack_lines(ll, encoded_buffer);
   return ret;
}

