#include "linelet_format.h"
#include "linelet.h"
#include "image.h"
#include "convert.h"
#include "memory.h"
#include "haar.h"
#include "timer.h"
#include "util.h"
#include "log.h"
#include "checksum.h"
#include "prepostfilter.h"

#include <pthread.h>
#include <stdlib.h>
#include <string.h>

#define min(a, b) ((a) < (b) ? (a) : (b))

struct thread_data
{
   uint8_t *buffer;
   size_t buffer_size;

   uint8_t *scanline_offset_buffer;
   uint32_t scanline_offset;

   unsigned width;
   unsigned start_y;
   unsigned height;
   unsigned real_height;

   unsigned slice_height;

   const uint8_t *src[LINELET_MAX_COMPONENTS];
   size_t stride[LINELET_MAX_COMPONENTS];

   struct linelet_yuv_image img[2]; // Need two slices for pre/post-filtering.

   size_t total_encoded;

   linelet_encoder_t *enc;
   pthread_t thread;
   pthread_mutex_t lock;
   pthread_cond_t cond;
   unsigned prefilter;
   bool ready;
   bool die;
   bool thread_active;

   void (*convert)(struct thread_data *thr, struct linelet_yuv_image *img, unsigned lines);
};

struct header_offsets
{
   size_t fmt1;
   size_t cod1;
   size_t off1;
   size_t scanline_offset;
   size_t eoh1;
   size_t payload;
};

struct linelet_format_encoder
{
   struct thread_data *threads;
   unsigned num_threads;
   unsigned real_height;

   struct linelet_format_encoder_options opts;
   struct linelet_component components[LINELET_MAX_COMPONENTS];
   unsigned num_components;

   void *buffer;
   size_t buffer_size;
   struct header_offsets offs;
};

static void prefilter_slices(struct linelet_yuv_image *a,
      struct linelet_yuv_image *b, unsigned slice_height, unsigned prefilter)
{
   for (unsigned c = 0; c < LINELET_MAX_COMPONENTS; c++)
   {
      unsigned stride = a->strides[c] >> 1;

      switch (prefilter)
      {
         case 3:
            linelet_prefilter_8(
                  a->components[c] + (slice_height - 4) * stride,
                  a->components[c] + (slice_height - 3) * stride,
                  a->components[c] + (slice_height - 2) * stride,
                  a->components[c] + (slice_height - 1) * stride,
                  b->components[c] + 0 * stride,
                  b->components[c] + 1 * stride,
                  b->components[c] + 2 * stride,
                  b->components[c] + 3 * stride,
                  stride);
            break;

         case 2:
            linelet_prefilter_4(
                  a->components[c] + (slice_height - 2) * stride,
                  a->components[c] + (slice_height - 1) * stride,
                  b->components[c] + 0 * stride,
                  b->components[c] + 1 * stride,
                  stride);
            break;

         case 1:
            linelet_prefilter_2(
                  a->components[c] + (slice_height - 1) * stride,
                  b->components[c] + 0 * stride,
                  stride);
            break;
      }
   }
}

static void convert_x2b10g10r10(struct thread_data *thr, struct linelet_yuv_image *img, unsigned lines)
{
   linelet_rct_forward_a2bgr10_yuv10(img->components,
         img->luma.stride,
         (const uint32_t*)thr->src[0], thr->width, lines, thr->stride[0]);
}

static void convert_x8r8g8b8(struct thread_data *thr, struct linelet_yuv_image *img, unsigned lines)
{
   linelet_rct_forward_argb8888(img->components,
         img->luma.stride,
         (const uint32_t*)thr->src[0], thr->width, lines, thr->stride[0]);
}

static void convert_gbrp10(struct thread_data *thr, struct linelet_yuv_image *img, unsigned lines)
{
   linelet_rct_forward_gbrp10(img->components,
         img->luma.stride,
         (const uint16_t**)thr->src, thr->width, lines, thr->stride);
}

static void convert_gbrp8(struct thread_data *thr, struct linelet_yuv_image *img, unsigned lines)
{
   linelet_rct_forward_gbrp8(img->components,
         img->luma.stride,
         (const uint8_t**)thr->src, thr->width, lines, thr->stride);
}

// This is in the format we want, just copy.
// The transforms mutate, so can't use original data directly.
static void convert_ycbcr_rct(struct thread_data *thr, struct linelet_yuv_image *img, unsigned lines)
{
   for (unsigned c = 0; c < 3; c++)
      for (unsigned y = 0; y < lines; y++)
         memcpy(img->components[c] + y * (img->strides[c] >> 1),
               thr->src[c] + y * thr->stride[c], img->strides[c]);
}

static void convert_ycbcr8_bt(struct thread_data *thr, struct linelet_yuv_image *img, unsigned lines)
{
   size_t widths[] = {
      img->width >> img->luma.sub,
      img->width >> img->cb.sub,
      img->width >> img->cr.sub,
   };

   linelet_bt_yuv8_forward_dc_shift(img->components, img->strides,
         (const uint8_t**)thr->src,
         widths,
         lines, thr->stride);
}

static void convert_ycbcr10_bt(struct thread_data *thr, struct linelet_yuv_image *img, unsigned lines)
{
   size_t widths[] = {
      img->width >> img->luma.sub,
      img->width >> img->cb.sub,
      img->width >> img->cr.sub,
   };

   linelet_bt_yuv10_forward_dc_shift(img->components, img->strides,
         (const uint16_t**)thr->src,
         widths,
         lines, thr->stride);
}

static void convert_ycbcr12_bt(struct thread_data *thr, struct linelet_yuv_image *img, unsigned lines)
{
   size_t widths[] = {
      img->width >> img->luma.sub,
      img->width >> img->cb.sub,
      img->width >> img->cr.sub,
   };

   linelet_bt_yuv12_forward_dc_shift(img->components, img->strides,
         (const uint16_t**)thr->src,
         widths,
         lines, thr->stride);
}

typedef void (*linelet_convert_func_t)(struct thread_data *thr, struct linelet_yuv_image *img, unsigned lines);
static linelet_convert_func_t format_to_convert_func(enum linelet_format_format fmt)
{
   switch (fmt)
   {
      case LINELET_FMT_X8R8G8B8:
         return convert_x8r8g8b8;
      case LINELET_FMT_X2B10G10R10:
         return convert_x2b10g10r10;
      case LINELET_FMT_GBRP8:
         return convert_gbrp8;
      case LINELET_FMT_GBRP10:
         return convert_gbrp10;

      case LINELET_FMT_YCBCR8_BT709_444:
      case LINELET_FMT_YCBCR8_BT709_422:
      case LINELET_FMT_YCBCR8_BT601_444:
      case LINELET_FMT_YCBCR8_BT601_422:
         return convert_ycbcr8_bt;

      case LINELET_FMT_YCBCR10_BT709_444:
      case LINELET_FMT_YCBCR10_BT709_422:
      case LINELET_FMT_YCBCR10_BT601_444:
      case LINELET_FMT_YCBCR10_BT601_422:
         return convert_ycbcr10_bt;

      case LINELET_FMT_YCBCR12_BT709_444:
      case LINELET_FMT_YCBCR12_BT709_422:
      case LINELET_FMT_YCBCR12_BT601_444:
      case LINELET_FMT_YCBCR12_BT601_422:
         return convert_ycbcr12_bt;

      case LINELET_FMT_YCBCR10_RCT_444:
      case LINELET_FMT_YCBCR10_RCT_422:
      case LINELET_FMT_YCBCR8_RCT_444:
      case LINELET_FMT_YCBCR8_RCT_422:
         return convert_ycbcr_rct;

      default:
         return NULL;
   }
}

static void *thread_func(void *data)
{
   struct thread_data *thr = data;

   for (;;)
   {
      pthread_mutex_lock(&thr->lock);
      while (!thr->ready && !thr->die)
         pthread_cond_wait(&thr->cond, &thr->lock);
      pthread_mutex_unlock(&thr->lock);

      if (thr->die)
         break;

      size_t target_size = thr->slice_height * thr->buffer_size / thr->height;

      if (target_size < sizeof(uint32_t))
         linelet_log_printf(LINELET_LOG_ERROR, "[Encoder]: Slice size too small.\n");
      else
      {
         target_size -= sizeof(uint32_t); // Checksum

         thr->total_encoded = 0;

         unsigned img_index = 0;
         unsigned height = thr->height;
         uint8_t *scanline_offset_buffer = thr->scanline_offset_buffer;

         if (thr->prefilter)
         {
            if (thr->start_y) // We have to redundantly convert a previous slice in order to be able to prefilter against it.
            {
               // Previous block.
               for (unsigned c = 0; c < LINELET_MAX_COMPONENTS; c++)
                  thr->src[c] -= thr->slice_height * thr->stride[c];
               // Previous blocks are always evenly sized.
               thr->convert(thr, &thr->img[img_index ^ 1], thr->slice_height);
               for (unsigned c = 0; c < LINELET_MAX_COMPONENTS; c++)
                  thr->src[c] += thr->slice_height * thr->stride[c];
            }

            // Current block.
            unsigned avail_y = LLMIN(thr->real_height - thr->start_y, thr->slice_height);
            thr->convert(thr, &thr->img[img_index], avail_y);
            for (unsigned c = 0; c < LINELET_MAX_COMPONENTS; c++)
               thr->src[c] += thr->slice_height * thr->stride[c];

            // Run prefilter against previous block.
            if (thr->start_y)
               prefilter_slices(&thr->img[img_index ^ 1], &thr->img[img_index], thr->slice_height, thr->prefilter);
         }

         for (unsigned y = 0; y < height; y += thr->slice_height)
         {
            // Convert in next block, apply prefilter on block edge.
            if (thr->prefilter)
            {
               unsigned next_real_y = thr->start_y + thr->slice_height + y;

               // Prefilter against next block if needed.
               if (next_real_y < thr->real_height)
               {
                  unsigned avail_y = LLMIN(thr->real_height - next_real_y, thr->slice_height);
                  thr->convert(thr, &thr->img[img_index ^ 1], avail_y);
                  prefilter_slices(&thr->img[img_index], &thr->img[img_index ^ 1], thr->slice_height, thr->prefilter);
               }
            }
            else
            {
               // Convert in current block, do not apply filter.
               unsigned real_y = thr->start_y + y;
               unsigned avail_y = LLMIN(thr->real_height - real_y, thr->slice_height);
               thr->convert(thr, &thr->img[img_index], avail_y);
            }

            size_t encoded = linelet_encoder_process(thr->enc,
                  thr->buffer + thr->total_encoded, target_size,
                  thr->img[img_index].components, thr->img[img_index].strides);

            TIMER_COUNTER(encoder_checksum);
            TIMER_START(encoder_checksum);
            linelet_write32_be(thr->buffer + thr->total_encoded + encoded,
                  linelet_checksum(thr->buffer + thr->total_encoded, encoded));
            TIMER_STOP(encoder_checksum);
            encoded += sizeof(uint32_t);

            linelet_write32_be(scanline_offset_buffer + 0, thr->total_encoded + thr->scanline_offset);
            linelet_write32_be(scanline_offset_buffer + sizeof(uint32_t), encoded);
            thr->total_encoded += encoded;

            scanline_offset_buffer += 2 * sizeof(uint32_t);

            for (unsigned c = 0; c < LINELET_MAX_COMPONENTS; c++)
               thr->src[c] += thr->slice_height * thr->stride[c];
            if (thr->prefilter)
               img_index ^= 1;
         }
      }

      pthread_mutex_lock(&thr->lock);
      thr->ready = false;
      pthread_cond_signal(&thr->cond);
      pthread_mutex_unlock(&thr->lock);
   }

   return NULL;
}

linelet_format_encoder_t *linelet_format_encoder_new(const struct linelet_format_encoder_options *opts)
{
   linelet_init();
   if (!opts->slice_height || (opts->slice_height & (opts->slice_height - 1)))
   {
      linelet_log_printf(LINELET_LOG_INFO, "[Encoder]: Intra interval must be positive and power-of-two.\n");
      return NULL;
   }

   linelet_format_encoder_t *enc = linelet_calloc(1, sizeof(*enc));
   if (!enc)
      return NULL;

   unsigned threads = opts->threads ? opts->threads : 1;
   enc->threads = linelet_calloc(threads, sizeof(*enc->threads));
   if (!enc->threads)
      goto error;
   enc->num_threads = threads;

   enc->real_height = opts->height;
   enc->opts = *opts;
   enc->opts.height = (enc->opts.height + opts->slice_height - 1) & ~(opts->slice_height - 1); // Round up so we have a whole number of slices.

   unsigned max_filter = linelet_log2i(opts->slice_height);
   if (enc->opts.filter > max_filter || enc->opts.filter > 3)
   {
      enc->opts.filter = LLMIN(max_filter, 3);
      linelet_log_printf(LINELET_LOG_WARN, "Filter value too high, capping to %u.\n", enc->opts.filter);
   }

   enc->num_components = linelet_format_to_components(opts->format);

   for (unsigned i = 0; i < 3; i++)
      enc->components[i].decomposition_levels = opts->decomposition_levels;
   enc->components[1].decimation = enc->components[2].decimation = linelet_format_to_subsample(opts->format);
   enc->components[1].decomposition_levels -= enc->components[1].decimation;
   enc->components[2].decomposition_levels -= enc->components[2].decimation;

   for (unsigned i = 0; i < threads; i++)
   {
      struct thread_data *data = &enc->threads[i];

      unsigned end_y = (enc->opts.height * (i + 1)) / threads;
      unsigned start_y = (enc->opts.height * i) / threads;

      // Round to intra intervals, except for last thread.
      if (i < threads - 1)
         end_y &= ~(opts->slice_height - 1);
      start_y &= ~(opts->slice_height - 1);

      data->width = opts->width;
      data->height = end_y - start_y;
      data->real_height = enc->real_height;
      data->start_y = start_y;
      data->slice_height = opts->slice_height;
      data->convert = format_to_convert_func(opts->format);
      data->prefilter = enc->opts.filter;

      for (unsigned j = 0; j < 2; j++)
      {
         if (!linelet_yuv_image_alloc(&data->img[j], opts->width, opts->slice_height,
                  enc->components[0].decimation,
                  enc->components[1].decimation,
                  enc->components[2].decimation))
            goto error;
      }

      data->enc = linelet_encoder_new(data->width, opts->slice_height,
            enc->components, 3,
            &(const struct linelet_encoder_options) {
               .precinct_samples = opts->precinct_samples,
               .rate_control = opts->rate_control,
               .q = opts->q,
            });

      if (!data->enc)
         goto error;

      data->thread_active = true;
      pthread_mutex_init(&data->lock, NULL);
      pthread_cond_init(&data->cond, NULL);
      pthread_create(&data->thread, NULL, thread_func, data);
   }

   return enc;
error:
   linelet_format_encoder_free(enc);
   return NULL;
}

void linelet_format_encoder_free(linelet_format_encoder_t *enc)
{
   if (!enc)
      return;

   for (unsigned i = 0; i < enc->num_threads; i++)
   {
      struct thread_data *data = &enc->threads[i];

      if (data->thread_active)
      {
         pthread_mutex_lock(&data->lock);
         data->ready = false;
         data->die = true;
         pthread_cond_signal(&data->cond);
         pthread_mutex_unlock(&data->lock);
         pthread_join(data->thread, NULL);
      }

      linelet_yuv_image_free(&data->img[0]);
      linelet_yuv_image_free(&data->img[1]);
      linelet_encoder_free(data->enc);
   }

   linelet_free(enc->threads);
   linelet_free(enc);
}

static void write_header(linelet_format_encoder_t *enc, uint8_t *buffer, const struct header_offsets *offs)
{
   memcpy(buffer, "LINELET1", 8);
   memcpy(buffer + offs->fmt1, "FMT1", 4);
   memcpy(buffer + offs->cod1, "COD1", 4);
   memcpy(buffer + offs->off1, "OFF1", 4);
   memcpy(buffer + offs->eoh1, "EOH1", 4);

   uint8_t *fmt1 = buffer + offs->fmt1;
   linelet_write32_be(fmt1 +  4, 24);
   linelet_write32_be(fmt1 +  8, enc->opts.width);
   linelet_write32_be(fmt1 + 12, enc->real_height);
   linelet_write16_be(fmt1 + 16, linelet_format_to_transform(enc->opts.format));
   linelet_write16_be(fmt1 + 18, linelet_format_to_bits(enc->opts.format));
   linelet_write16_be(fmt1 + 20, enc->opts.sample_aspect_ratio.num ? enc->opts.sample_aspect_ratio.num : 1);
   linelet_write16_be(fmt1 + 22, enc->opts.sample_aspect_ratio.den ? enc->opts.sample_aspect_ratio.den : 1);
   linelet_write32_be(fmt1 + 24, enc->opts.frame_rate.num ? enc->opts.frame_rate.num : 24000);
   linelet_write32_be(fmt1 + 28, enc->opts.frame_rate.den ? enc->opts.frame_rate.den : 1001);

   uint8_t *cod1 = buffer + offs->cod1;
   linelet_write32_be(cod1 + 4, 3 + 3 * 2);
   cod1[8] = linelet_log2i(enc->opts.precinct_samples);
   cod1[9] = linelet_log2i(enc->opts.slice_height);
   cod1[10] = 3 | (enc->opts.filter << 4);
   for (unsigned i = 0; i < 3; i++)
   {
      cod1[11 + 2 * i] = enc->components[i].decimation;
      cod1[12 + 2 * i] = enc->components[i].decomposition_levels;
   }

   linelet_write32_be(buffer + offs->off1 + 4, 2 * (enc->opts.height / enc->opts.slice_height) * sizeof(uint32_t));

   // Cannot write header checksum until we know slice size/offsets.
}

static struct header_offsets compute_header_offsets(linelet_format_encoder_t *enc)
{
   struct header_offsets offsets = {0};

   size_t off = 0;

   off += 8; // Magic (LINELET1)

   offsets.fmt1 = off;
   off += 4; // FMT1
   off += 4; // FMT1 len
   off += 4; //  Width
   off += 4; //  Height 
   off += 2; //  Color transform 
   off += 2; //  Bits
   off += 2; //  PAR num
   off += 2; //  PAR denom
   off += 4; //  FPS num
   off += 4; //  FPS den

   offsets.cod1 = off;
   off += 4; // COD1
   off += 4; // COD1 len
   off += 1; // Precinct samples
   off += 1; // Y slice height 
   off += 1; // Components

   for (unsigned i = 0; i < LINELET_MAX_COMPONENTS; i++)
   {
      off += 1; // Decimation
      off += 1; // Decomposition levels
   }

   offsets.off1 = off;
   off += 4; // OFF1
   off += 4; // OFF1 len

   offsets.scanline_offset = off;
   off += 2 * (enc->opts.height / enc->opts.slice_height) * sizeof(uint32_t);
   // Every scanline, offset in payload, size of scanline payload.

   offsets.eoh1 = off;
   off += 4; // EOH1
   off += 4; // CRC checksum
   offsets.payload = off;

   return offsets;
}

bool linelet_format_encoder_process_frame_async(linelet_format_encoder_t *enc, const void * const *components,
      const size_t *strides, void *buffer_, size_t buffer_size)
{
   uint8_t *buffer = buffer_;
   struct header_offsets offs = compute_header_offsets(enc);

   if (offs.payload >= buffer_size)
      return false;

   write_header(enc, buffer, &offs);

   size_t payload_size = buffer_size - offs.payload;

   for (unsigned i = 0; i < enc->num_threads; i++)
   {
      struct thread_data *data = &enc->threads[i];

      size_t buffer_start = (payload_size * i) / enc->num_threads;
      size_t buffer_end = (payload_size * (i + 1)) / enc->num_threads;

      for (unsigned c = 0; c < enc->num_components; c++)
      {
         data->src[c] = (const uint8_t*)components[c] + data->start_y * strides[c];
         data->stride[c] = strides[c];
      }

      data->buffer_size = buffer_end - buffer_start;
      data->buffer = buffer + offs.payload + buffer_start;
      data->scanline_offset = buffer_start;
      data->scanline_offset_buffer = buffer + offs.scanline_offset + (data->start_y / enc->opts.slice_height) * 2 * sizeof(uint32_t);

      pthread_mutex_lock(&data->lock);
      data->ready = true;
      pthread_cond_signal(&data->cond);
      pthread_mutex_unlock(&data->lock);
   }

   enc->buffer = buffer;
   enc->buffer_size = buffer_size;
   enc->offs = offs;
   return true;
}

size_t linelet_format_encoder_process_frame(linelet_format_encoder_t *enc, const void * const *components,
      const size_t *strides, void *buffer, size_t buffer_size)
{
   if (linelet_format_encoder_process_frame_async(enc, components, strides, buffer, buffer_size))
      return linelet_format_encoder_process_frame_wait(enc);
   else
      return 0;
}

size_t linelet_format_encoder_process_frame_wait(linelet_format_encoder_t *enc)
{
   size_t total_encoded = 0;
   // Wait until all threads are done processing.
   for (unsigned i = 0; i < enc->num_threads; i++)
   {
      struct thread_data *data = &enc->threads[i];
      pthread_mutex_lock(&data->lock);
      while (data->ready && !data->die)
         pthread_cond_wait(&data->cond, &data->lock);
      pthread_mutex_unlock(&data->lock);
      total_encoded += data->total_encoded;
   }

   TIMER_COUNTER(pack_frame);
   TIMER_START_COUNT(pack_frame, enc->buffer_size);
   if (enc->num_threads > 1)
   {
      // Tightly pack the payload. If we use multiple encoding threads, the payload is scattered around.
      for (unsigned i = 1; i < enc->num_threads; i++)
      {
         struct thread_data *prev = &enc->threads[i - 1];
         struct thread_data *cur = &enc->threads[i];

         uint32_t prev_thread_end = prev->scanline_offset + prev->total_encoded;
         uint32_t thread_start = cur->scanline_offset;
         uint32_t offset = thread_start - prev_thread_end;

         memmove(cur->buffer - offset, cur->buffer, cur->total_encoded);

         // Readjust slice offsets.
         unsigned slices = cur->height / enc->opts.slice_height;
         uint8_t *buf = cur->scanline_offset_buffer;
         for (unsigned s = 0; s < slices; s++, buf += 2 * sizeof(uint32_t))
            linelet_write32_be(buf, linelet_read32_be(buf) - offset);

         cur->scanline_offset = prev_thread_end;
      }
   }
   TIMER_STOP(pack_frame);

   // Header checksum.
   linelet_write32_be(enc->buffer + enc->offs.eoh1 + 4, linelet_checksum(enc->buffer, enc->offs.eoh1));

   return enc->threads[enc->num_threads - 1].scanline_offset + enc->threads[enc->num_threads - 1].total_encoded + enc->offs.payload;
}

