#include <GLFW/glfw3.h>
#include <stdlib.h>
#include <stdint.h>
#include <stdio.h>
#include <stdbool.h>
#include <string.h>
#include <math.h>
#include "linelet_format.h"
#include "memory.h"
#include "timer.h"
#include "util.h"
#include "log.h"

#include "glsym/glsym.h"

static const char vertex_program[] =
   "#version 120\n"
   "attribute vec2 aVertex;\n"
   "attribute vec2 aTexCoord;\n"
   "varying vec2 vTex;\n"
   "void main() { gl_Position = vec4(aVertex, 0.0, 1.0); vTex = aTexCoord; }\n";

static const char fragment_program_single[] =
   "#version 120\n"
   "varying vec2 vTex;\n"
   "uniform sampler2D sTexture[3];\n"
   "uniform float uScale;\n"
   "uniform mat3 uTransform;\n"
   "void main() {\n"
   "   float y = texture2D(sTexture[0], vTex).r;\n"
   "   float cb = texture2D(sTexture[1], vTex).r;\n"
   "   float cr = texture2D(sTexture[2], vTex).r;\n"
   "   vec3 yuv = uScale * vec3(y, cb, cr);\n"
   "   yuv -= vec3(0.0625, 0.5, 0.5);\n"
   "   vec3 rgb = uTransform * yuv;\n"
   "   gl_FragColor = vec4(rgb, 1.0);\n"
   "}";

static const char fragment_program_error[] =
   "#version 120\n"
   "varying vec2 vTex;\n"
   "uniform sampler2D sTexture[3];\n"
   "uniform sampler2D sTextureRef[3];\n"
   "uniform float uScale;\n"
   "uniform mat3 uTransform;\n"
   "void main() {\n"

   "   float y = texture2D(sTexture[0], vTex).r;\n"
   "   float cb = texture2D(sTexture[1], vTex).r;\n"
   "   float cr = texture2D(sTexture[2], vTex).r;\n"
   "   vec3 yuv = uScale * vec3(y, cb, cr);\n"
   "   yuv -= vec3(0.0625, 0.5, 0.5);\n"
   "   vec3 rgb = uTransform * yuv;\n"

   "   y = texture2D(sTextureRef[0], vTex).r;\n"
   "   cb = texture2D(sTextureRef[1], vTex).r;\n"
   "   cr = texture2D(sTextureRef[2], vTex).r;\n"
   "   yuv = uScale * vec3(y, cb, cr);\n"
   "   yuv -= vec3(0.0625, 0.5, 0.5);\n"
   "   vec3 rgb_ref = uTransform * yuv;\n"

   "   gl_FragColor = vec4(0.5 + 32.0 * (rgb - rgb_ref), 1.0);\n"
   "}";

static const char fragment_program_split[] =
   "#version 120\n"
   "varying vec2 vTex;\n"
   "uniform sampler2D sTexture[3];\n"
   "uniform sampler2D sTextureRef[3];\n"
   "uniform float uScale;\n"
   "uniform mat3 uTransform;\n"
   "void main() {\n"

   "   float y;\n"
   "   float cb;\n"
   "   float cr;\n"
   "   if (vTex.x > 0.5) {\n"
   "      y = texture2D(sTexture[0], vTex).r;\n"
   "      cb = texture2D(sTexture[1], vTex).r;\n"
   "      cr = texture2D(sTexture[2], vTex).r;\n"
   "   } else {\n"
   "      y = texture2D(sTextureRef[0], vTex).r;\n"
   "      cb = texture2D(sTextureRef[1], vTex).r;\n"
   "      cr = texture2D(sTextureRef[2], vTex).r;\n"
   "   }\n"

   "   vec3 yuv = uScale * vec3(y, cb, cr);\n"
   "   yuv -= vec3(0.0625, 0.5, 0.5);\n"
   "   vec3 rgb = uTransform * yuv;\n"

   "   gl_FragColor = vec4(rgb, 1.0);\n"
   "}";


static GLuint shader_prog_single;
static GLuint shader_prog_error;
static GLuint shader_prog_split;
static GLuint shader_prog; // Current mode

static GLuint pbo;
static GLuint vbo;
static unsigned last_width, last_height;
static unsigned last_width_ref, last_height_ref;
static GLuint tex[6];

static GLuint compile_shader(GLenum type, const char *source)
{
   GLuint shader = glCreateShader(type);
   glShaderSource(shader, 1, (const GLchar**)&source, NULL);
   glCompileShader(shader);
   
   GLint status = 0;
   glGetShaderiv(shader, GL_COMPILE_STATUS, &status);

   if (!status)
   {
      linelet_log_printf(LINELET_LOG_INFO, "Failed to compile.\n");
      char log_info[8 * 1024];
      GLsizei log_len;
      glGetShaderInfoLog(shader, sizeof(log_info), &log_len, log_info);
      linelet_log_printf(LINELET_LOG_INFO, "ERROR: %s\n", log_info);
      return 0;
   }

   return shader;
}

static GLuint compile_program(const char *vertex_source, const char *fragment_source)
{
   GLuint prog = glCreateProgram();
   GLuint vert = compile_shader(GL_VERTEX_SHADER, vertex_source);
   GLuint frag = compile_shader(GL_FRAGMENT_SHADER, fragment_source);

   glAttachShader(prog, vert);
   glAttachShader(prog, frag);

   glBindAttribLocation(prog, 0, "aVertex");
   glBindAttribLocation(prog, 1, "aTexCoord");
   glLinkProgram(prog);

   GLint status = 0;
   glGetProgramiv(prog, GL_LINK_STATUS, &status);
   if (!status)
   {
      linelet_log_printf(LINELET_LOG_INFO, "Failed to link.\n");
      char log_info[8 * 1024];
      GLsizei log_len;
      glGetProgramInfoLog(prog, sizeof(log_info), &log_len, log_info);
      linelet_log_printf(LINELET_LOG_INFO, "ERROR: %s\n", log_info);
      exit(EXIT_FAILURE);
   }

   glDeleteShader(vert);
   glDeleteShader(frag);

   return prog;
}

static const float *format_to_transform(enum linelet_format_format fmt)
{
   static const float bt709[9] = {
      1.164, 1.164, 1.164,
      0.0, -0.213, 2.112,
      1.793, -0.533, 0.0,
   };

   static const float bt601[9] = {
      1.164, 1.164, 1.164,
      0.0, -0.392, 2.017,
      1.596, -0.813, 0.0,
   };

   switch (fmt)
   {
      case LINELET_FMT_YCBCR10_BT709_422:
      case LINELET_FMT_YCBCR8_BT709_422:
      case LINELET_FMT_YCBCR10_BT709_444:
      case LINELET_FMT_YCBCR8_BT709_444:
      default:
         return bt709;

      case LINELET_FMT_YCBCR10_BT601_422:
      case LINELET_FMT_YCBCR8_BT601_422:
      case LINELET_FMT_YCBCR10_BT601_444:
      case LINELET_FMT_YCBCR8_BT601_444:
         return bt601;
   }
}

static int64_t total_nsec;
static unsigned total_frames;
static linelet_format_decoder_t *dec;

static void set_viewport(unsigned vp_width, unsigned vp_height,
      unsigned src_width, unsigned src_height)
{
   float src_aspect = (float)src_width / src_height;
   float dst_aspect = (float)vp_width / vp_height;

   int x = 0; 
   int y = 0;
   int w = vp_width;
   int h = vp_height;

   if (fabsf(dst_aspect - src_aspect) < 0.0001f)
   {
      // Skip
   }
   else if (dst_aspect > src_aspect)
   {
      float delta = (src_aspect / dst_aspect - 1.0f) / 2.0f + 0.5f;
      x = (int)roundf(vp_width * (0.5f - delta));
      w = (int)roundf(2.0f * vp_width * delta);
   }
   else
   {
      float delta = (dst_aspect / src_aspect - 1.0f) / 2.0f + 0.5f;
      y = (int)roundf(vp_height * (0.5f - delta));
      h = (int)roundf(2.0f * vp_height * delta);
   }

   glViewport(x, y, w, h);
}

static bool decode(FILE *file, unsigned tex_unit, struct linelet_format_decoder_frame_params *out_params,
      unsigned *tex_width, unsigned *tex_height)
{
   uint8_t frame_size_buf[4];
   if (fread(frame_size_buf, sizeof(frame_size_buf), 1, file) != 1)
      return false;
   uint32_t frame_size = linelet_read32_be(frame_size_buf);

   void *encoded_buffer = linelet_malloc(frame_size);
   if (!encoded_buffer)
      return false;

   if (fread(encoded_buffer, 1, frame_size, file) != frame_size)
      return false;

   struct linelet_format_decoder_frame_params params = {0};
   if (!linelet_format_decoder_parse_header(&params, encoded_buffer, frame_size))
      return false;

   if (!dec)
      dec = linelet_format_decoder_new(&(const struct linelet_format_decoder_options) { .format = params.format, .threads = 8 });
   if (!dec)
      return false;

   glBindBuffer(GL_PIXEL_UNPACK_BUFFER, 0);

   unsigned subsample = linelet_format_to_subsample(params.format);
   unsigned bpp = linelet_format_to_bpp(params.format);
   unsigned bits = linelet_format_to_bits(params.format);
   if (!bpp)
      return false;

   size_t luma_size = params.width * params.height * bpp;
   size_t chroma_size = luma_size >> subsample;

   GLenum internal_format = bpp == 2 ? GL_R16 : GL_R8;
   GLenum type = bpp == 2 ? GL_UNSIGNED_SHORT : GL_UNSIGNED_BYTE;

   GLint scale_loc = glGetUniformLocation(shader_prog, "uScale");
   GLfloat scale;
   if (bits == 12)
      scale = (float)0xffff / 4095;
   else if (bits == 10)
      scale = (float)0xffff / 1023;
   else
      scale = 1.0f;
   glUniform1f(scale_loc, scale);

   glUniformMatrix3fv(glGetUniformLocation(shader_prog, "uTransform"),
         1, GL_FALSE, format_to_transform(params.format));

   if (params.width != *tex_width || params.height != *tex_height)
   {
      *tex_width = params.width;
      *tex_height = params.height;

      for (unsigned i = 0; i < 3; i++)
      {
         glBindTexture(GL_TEXTURE_2D, tex[tex_unit + i]);
         glTexImage2D(GL_TEXTURE_2D, 0, internal_format,
               params.width >> (i ? subsample : 0), params.height,
               0, GL_RED, type, NULL);
      }
   }

   size_t pbo_size = luma_size + chroma_size * 2;
   glBindBuffer(GL_PIXEL_UNPACK_BUFFER, pbo);
   glBufferData(GL_PIXEL_UNPACK_BUFFER, pbo_size,
         NULL, GL_STREAM_DRAW);

   uint8_t *buf = glMapBufferRange(GL_PIXEL_UNPACK_BUFFER, 0, pbo_size,
         GL_MAP_WRITE_BIT | GL_MAP_INVALIDATE_BUFFER_BIT);

   void *dsts[] = { buf, buf + luma_size, buf + luma_size + chroma_size };
   const size_t strides[] = { params.width * bpp, (params.width >> subsample) * bpp, (params.width >> subsample) * bpp };
   uintptr_t offset[] = { 0, luma_size, luma_size + chroma_size };

   bool decoded = false;
   if (buf)
   {
      int64_t start = linelet_timer_get_nsec();
      decoded = linelet_format_decoder_process_frame(dec,
            encoded_buffer, frame_size, dsts, strides);
      total_nsec += linelet_timer_get_nsec() - start;
      total_frames++;
      if (!decoded)
         linelet_log_printf(LINELET_LOG_INFO, "Failed to decode frame.\n");
   }
   glUnmapBuffer(GL_PIXEL_UNPACK_BUFFER);

   linelet_log_printf(LINELET_LOG_INFO, "\r[Viewer]: Decoder FPS: %8.3f, Decoding Time: %8.3f s",
         1000000000.0 * total_frames / total_nsec,
         total_nsec / 1000000000.0);

   for (unsigned i = 0; i < 3; i++)
   {
      glActiveTexture(GL_TEXTURE0 + i + tex_unit);
      glBindTexture(GL_TEXTURE_2D, tex[tex_unit + i]);
      glTexSubImage2D(GL_TEXTURE_2D, 0, 0, 0,
            params.width >> (i ? subsample : 0), params.height,
            GL_RED, type, (const void*)offset[i]);
      glGenerateMipmap(GL_TEXTURE_2D);
   }
   glActiveTexture(GL_TEXTURE0);

   linelet_free(encoded_buffer);
   *out_params = params;
   return decoded;
}

static bool render(FILE *file, FILE *file_ref, unsigned width, unsigned height)
{
   glUseProgram(shader_prog);

   struct linelet_format_decoder_frame_params params = {0};
   if (!decode(file, 0, &params, &last_width, &last_height))
      return false;

   if (file_ref && !decode(file_ref, 3, &params, &last_width_ref, &last_height_ref))
      return false;

   set_viewport(width, height, params.width, params.height);

   glBindBuffer(GL_ARRAY_BUFFER, vbo);
   glEnableVertexAttribArray(0);
   glEnableVertexAttribArray(1);
   glVertexAttribPointer(0, 2, GL_BYTE, GL_FALSE, 0, (const GLvoid*)0);
   glVertexAttribPointer(1, 2, GL_BYTE, GL_FALSE, 0, (const GLvoid*)8);

   glDrawArrays(GL_TRIANGLE_STRIP, 0, 4);

   return true;
}

static void init(void)
{
   shader_prog_single = compile_program(vertex_program, fragment_program_single);
   shader_prog_error = compile_program(vertex_program, fragment_program_error);
   shader_prog_split = compile_program(vertex_program, fragment_program_split);

   shader_prog = shader_prog_single;

   glUseProgram(shader_prog_single);
   glUniform1iv(glGetUniformLocation(shader_prog_single, "sTexture"), 3, (GLint[]) {0, 1, 2});
   glUseProgram(shader_prog_error);
   glUniform1iv(glGetUniformLocation(shader_prog_error, "sTexture"), 3, (GLint[]) {0, 1, 2});
   glUniform1iv(glGetUniformLocation(shader_prog_error, "sTextureRef"), 3, (GLint[]) {3, 4, 5});
   glUseProgram(shader_prog_split);
   glUniform1iv(glGetUniformLocation(shader_prog_split, "sTexture"), 3, (GLint[]) {0, 1, 2});
   glUniform1iv(glGetUniformLocation(shader_prog_split, "sTextureRef"), 3, (GLint[]) {3, 4, 5});

   glGenTextures(6, tex);

   for (unsigned i = 0; i < 6; i++)
   {
      glBindTexture(GL_TEXTURE_2D, tex[i]);
      glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
      glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR);
      glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
      glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
   }

   glGenBuffers(1, &pbo);

   glGenBuffers(1, &vbo);
   const int8_t vbo_data[] = {
      -1, -1, 1, -1, -1, 1, 1, 1, // Vertices
       0,  1, 1,  1,  0, 0, 1, 0, // TexCoord
   };

   glBindBuffer(GL_ARRAY_BUFFER, vbo);
   glBufferData(GL_ARRAY_BUFFER, sizeof(vbo_data), vbo_data, GL_STATIC_DRAW);
}

static bool die;
static void key_cb(GLFWwindow *window, int key, int scancode, int action, int mods)
{
   (void)window;
   (void)scancode;
   (void)mods;
   if (key == GLFW_KEY_ESCAPE && action == GLFW_PRESS)
      die = true;
   if (key == GLFW_KEY_S && action == GLFW_PRESS)
   {
      shader_prog = shader_prog_single;
      glfwSetWindowTitle(window, "LLV viewer - Single");
   }
   else if (key == GLFW_KEY_E && action == GLFW_PRESS)
   {
      shader_prog = shader_prog_error;
      glfwSetWindowTitle(window, "LLV viewer - Error");
   }

   else if (key == GLFW_KEY_P && action == GLFW_PRESS)
   {
      shader_prog = shader_prog_split;
      glfwSetWindowTitle(window, "LLV viewer - Split");
   }
}

static bool check_magic(FILE *file)
{
   char magic[4];
   if (fread(&magic, sizeof(magic), 1, file) != 1)
      return false;

   return memcmp(magic, "LLV1", 4) == 0;
}

int main(int argc, char *argv[])
{
   if (argc < 2)
      return EXIT_FAILURE;

   FILE *raw_file = strcmp(argv[1], "-") ? fopen(argv[1], "rb") : stdin;
   if (!raw_file)
   {
      linelet_log_printf(LINELET_LOG_INFO, "Failed to open raw file.\n");
      return EXIT_FAILURE;
   }

   FILE *raw_file_ref = argc >= 3 ? fopen(argv[2], "rb") : NULL;

   if (!check_magic(raw_file) || (raw_file_ref && !check_magic(raw_file_ref)))
   {
      linelet_log_printf(LINELET_LOG_INFO, "Failed to read raw file.\n");
      return EXIT_FAILURE;
   }

   if (!glfwInit())
      return EXIT_FAILURE;

   GLFWmonitor *monitor = NULL;
#if 0
   int monitor_count = 0;
   GLFWmonitor **monitors = glfwGetMonitors(&monitor_count);
   int max_pixels = 0;
   for (int i = 0; i < monitor_count; i++)
   {
      int w = 0, h = 0;
      glfwGetMonitorPhysicalSize(monitors[i], &w, &h);
      w *= h;
      if (w > max_pixels)
      {
         monitor = monitors[i];
         max_pixels = w;
      }
   }
#endif

   int target_width = monitor ? glfwGetVideoMode(monitor)->width : 1280;
   int target_height = monitor ? glfwGetVideoMode(monitor)->height : 720;

   glfwWindowHint(GLFW_DEPTH_BITS, 0);
   glfwWindowHint(GLFW_STENCIL_BITS, 0);

   GLFWwindow *window = glfwCreateWindow(target_width, target_height,
         "LLV viewer - Single", monitor, NULL);

   if (!window)
   {
      glfwTerminate();
      return EXIT_FAILURE; 
   }

   glfwMakeContextCurrent(window);

   rglgen_resolve_symbols((rglgen_proc_address_t)glfwGetProcAddress);
   glfwSwapInterval(1);

   init();
   glfwSetKeyCallback(window, key_cb);

   while (!die && !glfwWindowShouldClose(window))
   {
      glClearColor(0.0f, 0.0f, 0.0f, 1.0f);
      glClear(GL_COLOR_BUFFER_BIT);

      int width = 0, height = 0;
      glfwGetFramebufferSize(window, &width, &height);

      if (!render(raw_file, raw_file_ref, width, height))
         break;

      glfwSwapBuffers(window);
      glfwPollEvents();
   }

   if (window)
      glfwDestroyWindow(window);
   glfwTerminate();

   if (raw_file && raw_file != stdin)
      fclose(raw_file);
   if (raw_file_ref)
      fclose(raw_file_ref);
   linelet_format_decoder_free(dec);
   linelet_timer_log();
}

